import csv
import os
import warnings
from pathlib import Path

import numpy as np
import pytest
import scipy.sparse
import xgboost as xgb
from hypothesis import given, settings, strategies
from scipy.sparse import csr_matrix, rand
from xgboost import testing as tm
from xgboost.core import DataSplitMode
from xgboost.testing.data import np_dtypes, run_base_margin_info
from xgboost.testing.utils import predictor_equal

dpath = "demo/data/"
rng = np.random.RandomState(1994)


class TestDMatrix:
    def test_warn_missing(self):
        from xgboost import data

        with pytest.warns(UserWarning):
            data._warn_unused_missing("uri", 4)

        with warnings.catch_warnings():
            warnings.simplefilter("error")
            data._warn_unused_missing("uri", None)
            data._warn_unused_missing("uri", np.nan)

        with warnings.catch_warnings():
            warnings.simplefilter("error")
            x = rng.randn(10, 10)
            y = rng.randn(10)

            xgb.DMatrix(x, y, missing=4)

    def test_dmatrix_numpy_init(self):
        data = np.random.randn(5, 5)
        dm = xgb.DMatrix(data)
        assert dm.num_row() == 5
        assert dm.num_col() == 5

        data = np.array([[1, 2], [3, 4]])
        dm = xgb.DMatrix(data)
        assert dm.num_row() == 2
        assert dm.num_col() == 2

        # 0d array
        with pytest.raises(ValueError):
            xgb.DMatrix(np.array(1))
        # 1d array
        with pytest.raises(ValueError):
            xgb.DMatrix(np.array([1, 2, 3]))
        # 3d array
        data = np.random.randn(5, 5, 5)
        with pytest.raises(ValueError):
            xgb.DMatrix(data)
        # object dtype
        data = np.array([["a", "b"], ["c", "d"]])
        with pytest.raises(ValueError):
            xgb.DMatrix(data)

    def test_np_view(self):
        # Sliced Float32 array
        y = np.array([12, 34, 56], np.float32)[::2]
        from_view = xgb.DMatrix(np.array([[]]), label=y).get_label()
        from_array = xgb.DMatrix(np.array([[]]), label=y + 0).get_label()
        assert from_view.shape == from_array.shape
        assert (from_view == from_array).all()

        # Sliced UInt array
        z = np.array([12, 34, 56], np.uint32)[::2]
        dmat = xgb.DMatrix(np.array([[]]))
        dmat.set_uint_info("group", z)
        from_view = dmat.get_uint_info("group_ptr")
        dmat = xgb.DMatrix(np.array([[]]))
        dmat.set_uint_info("group", z + 0)
        from_array = dmat.get_uint_info("group_ptr")
        assert from_view.shape == from_array.shape
        assert (from_view == from_array).all()

    def test_slice(self):
        X = rng.randn(100, 100)
        y = rng.randint(low=0, high=3, size=100).astype(np.float32)
        d = xgb.DMatrix(X, y)
        np.testing.assert_equal(d.get_label(), y)

        fw = rng.uniform(size=100).astype(np.float32)
        d.set_info(feature_weights=fw)

        # base margin is per-class in multi-class classifier
        base_margin = rng.randn(100, 3).astype(np.float32)
        d.set_base_margin(base_margin)
        np.testing.assert_allclose(d.get_base_margin().reshape(100, 3), base_margin)

        ridxs = [1, 2, 3, 4, 5, 6]
        sliced = d.slice(ridxs)

        # Slicing works with label and other meta info fields
        np.testing.assert_equal(sliced.get_label(), y[1:7])
        np.testing.assert_equal(sliced.get_float_info("feature_weights"), fw)
        np.testing.assert_equal(sliced.get_base_margin(), base_margin[1:7, :])

        # Slicing a DMatrix results into a DMatrix that's equivalent to a DMatrix that's
        # constructed from the corresponding NumPy slice
        d2 = xgb.DMatrix(X[1:7, :], y[1:7])
        d2.set_base_margin(base_margin[1:7, :])
        eval_res = {}
        _ = xgb.train(
            {"num_class": 3, "objective": "multi:softprob", "eval_metric": "mlogloss"},
            d,
            num_boost_round=2,
            evals=[(d2, "d2"), (sliced, "sliced")],
            evals_result=eval_res,
        )
        np.testing.assert_equal(
            eval_res["d2"]["mlogloss"], eval_res["sliced"]["mlogloss"]
        )

        ridxs_arr = np.array(ridxs)[1:]  # handles numpy slice correctly
        sliced = d.slice(ridxs_arr)
        np.testing.assert_equal(sliced.get_label(), y[2:7])

    def test_feature_names_slice(self):
        data = np.random.randn(5, 5)

        # different length
        with pytest.raises(ValueError):
            xgb.DMatrix(data, feature_names=list("abcdef"))
        # contains duplicates
        with pytest.raises(ValueError):
            xgb.DMatrix(data, feature_names=["a", "b", "c", "d", "d"])
        # contains symbol
        with pytest.raises(ValueError):
            xgb.DMatrix(data, feature_names=["a", "b", "c", "d", "e<1"])

        dm = xgb.DMatrix(data)
        dm.feature_names = list("abcde")
        assert dm.feature_names == list("abcde")

        assert dm.slice([0, 1]).num_col() == dm.num_col()
        assert dm.slice([0, 1]).feature_names == dm.feature_names

        with pytest.raises(ValueError, match=r"Duplicates found: \[.*'bar'.*\]"):
            dm.feature_names = ["bar"] * (data.shape[1] - 2) + ["a", "b"]

        dm.feature_types = list("qiqiq")
        assert dm.feature_types == list("qiqiq")

        with pytest.raises(ValueError):
            dm.feature_types = list("abcde")

        # reset
        dm.feature_names = None
        dm.feature_types = None
        assert dm.feature_names is None
        assert dm.feature_types is None

    def test_feature_names(self):
        data = np.random.randn(100, 5)
        target = np.array([0, 1] * 50)

        cases = [
            ["Feature1", "Feature2", "Feature3", "Feature4", "Feature5"],
            ["要因1", "要因2", "要因3", "要因4", "要因5"],
        ]

        for features in cases:
            dm = xgb.DMatrix(data, label=target, feature_names=features)
            assert dm.feature_names == features
            assert dm.num_row() == 100
            assert dm.num_col() == 5

            params = {
                "objective": "multi:softprob",
                "eval_metric": "mlogloss",
                "eta": 0.3,
                "num_class": 3,
            }

            bst = xgb.train(params, dm, num_boost_round=10)
            scores = bst.get_fscore()
            assert list(sorted(k for k in scores)) == features

            dummy = np.random.randn(5, 5)
            dm = xgb.DMatrix(dummy, feature_names=features)
            bst.predict(dm)

            # different feature name must raises error
            dm = xgb.DMatrix(dummy, feature_names=list("abcde"))
            with pytest.raises(ValueError):
                bst.predict(dm)

    @pytest.mark.skipif(**tm.no_pandas())
    def test_save_binary(self, tmp_path: Path) -> None:
        import pandas as pd

        path = tmp_path / "m.dmatrix"
        data = pd.DataFrame({"a": [0, 1], "b": [2, 3], "c": [4, 5]})
        m0 = xgb.DMatrix(data.loc[:, ["a", "b"]], data["c"])
        assert m0.feature_names == ["a", "b"]
        m0.save_binary(path)
        m1 = xgb.DMatrix(path)
        assert m0.feature_names == m1.feature_names
        assert m0.feature_types == m1.feature_types

    def test_get_info(self):
        dtrain, _ = tm.load_agaricus(__file__)
        dtrain.get_float_info("label")
        dtrain.get_float_info("weight")
        dtrain.get_float_info("base_margin")
        dtrain.get_uint_info("group_ptr")

        group_len = np.array([2, 3, 4])
        dtrain.set_group(group_len)
        np.testing.assert_equal(group_len, dtrain.get_group())

    def test_qid(self):
        rows = 100
        cols = 10
        X, y = rng.randn(rows, cols), rng.randn(rows)
        qid = rng.randint(low=0, high=10, size=rows, dtype=np.uint32)
        qid = np.sort(qid)

        Xy = xgb.DMatrix(X, y)
        Xy.set_info(qid=qid)
        group_ptr = Xy.get_uint_info("group_ptr")
        assert group_ptr[0] == 0
        assert group_ptr[-1] == rows

    def test_feature_weights(self):
        kRows = 10
        kCols = 50
        rng = np.random.RandomState(1994)
        fw = rng.uniform(size=kCols)
        X = rng.randn(kRows, kCols)
        m = xgb.DMatrix(X)
        m.set_info(feature_weights=fw)
        np.testing.assert_allclose(fw, m.get_float_info("feature_weights"))
        # Handle empty
        m.set_info(feature_weights=np.empty((0,)))

        assert m.get_float_info("feature_weights").shape[0] == 0

        fw -= 1

        with pytest.raises(ValueError):
            m.set_info(feature_weights=fw)

    def test_sparse_dmatrix_csr(self, tmp_path: Path) -> None:
        nrow = 100
        ncol = 1000
        x = rand(nrow, ncol, density=0.0005, format="csr", random_state=rng)
        assert x.indices.max() < ncol
        x.data[:] = 1
        dtrain = xgb.DMatrix(x, label=rng.binomial(1, 0.3, nrow))
        assert (dtrain.num_row(), dtrain.num_col()) == (nrow, ncol)
        watchlist = [(dtrain, "train")]
        param = {"max_depth": 3, "objective": "binary:logistic"}
        bst = xgb.train(param, dtrain, 5, evals=watchlist)
        bst.predict(dtrain)

        i32 = csr_matrix((x.data.astype(np.int32), x.indices, x.indptr), shape=x.shape)
        f32 = csr_matrix(
            (i32.data.astype(np.float32), x.indices, x.indptr), shape=x.shape
        )
        di32 = xgb.DMatrix(i32)
        df32 = xgb.DMatrix(f32)
        dense = xgb.DMatrix(f32.toarray(), missing=0)

        path = tmp_path / "f32.dmatrix"
        df32.save_binary(path)
        with open(path, "rb") as fd:
            df32_buffer = np.array(fd.read())
        path = tmp_path / "i32.dmatrix"
        di32.save_binary(path)
        with open(path, "rb") as fd:
            di32_buffer = np.array(fd.read())

        path = tmp_path / "dense.dmatrix"
        dense.save_binary(path)
        with open(path, "rb") as fd:
            dense_buffer = np.array(fd.read())

        np.testing.assert_equal(df32_buffer, di32_buffer)
        np.testing.assert_equal(df32_buffer, dense_buffer)

    def test_sparse_dmatrix_csc(self):
        nrow = 1000
        ncol = 100
        x = rand(nrow, ncol, density=0.0005, format="csc", random_state=rng)
        assert x.indices.max() < nrow - 1
        x.data[:] = 1
        dtrain = xgb.DMatrix(x, label=rng.binomial(1, 0.3, nrow))
        assert (dtrain.num_row(), dtrain.num_col()) == (nrow, ncol)
        watchlist = [(dtrain, "train")]
        param = {"max_depth": 3, "objective": "binary:logistic"}
        bst = xgb.train(param, dtrain, 5, evals=watchlist)
        bst.predict(dtrain)

    def test_unknown_data(self):
        class Data:
            pass

        with pytest.raises(TypeError):
            with pytest.warns(UserWarning):
                d = Data()
                xgb.DMatrix(d)

        from scipy import sparse

        rng = np.random.RandomState(1994)
        X = rng.rand(10, 10)
        y = rng.rand(10)
        X = sparse.dok_matrix(X)
        with pytest.warns(UserWarning, match="dok_matrix"):
            Xy = xgb.DMatrix(X, y)
            assert Xy.num_row() == 10
            assert Xy.num_col() == 10

    @pytest.mark.skipif(**tm.no_pandas())
    def test_np_categorical(self):
        n_features = 10
        X, y = tm.make_categorical(10, n_features, n_categories=4, onehot=False)
        X = X.values.astype(np.float32)
        feature_types = ["c"] * n_features

        assert isinstance(X, np.ndarray)
        Xy = xgb.DMatrix(X, y, feature_types=feature_types)
        np.testing.assert_equal(np.array(Xy.feature_types), np.array(feature_types))

    def test_scipy_categorical(self):
        from scipy import sparse

        n_features = 10
        X, y = tm.make_categorical(10, n_features, n_categories=4, onehot=False)
        X = X.values.astype(np.float32)
        feature_types = ["c"] * n_features

        X[1, 3] = np.nan
        X[2, 4] = np.nan
        X = sparse.csr_matrix(X)

        Xy = xgb.DMatrix(X, y, feature_types=feature_types)
        np.testing.assert_equal(np.array(Xy.feature_types), np.array(feature_types))

        X = sparse.csc_matrix(X)

        Xy = xgb.DMatrix(X, y, feature_types=feature_types)
        np.testing.assert_equal(np.array(Xy.feature_types), np.array(feature_types))

        X = sparse.coo_matrix(X)

        Xy = xgb.DMatrix(X, y, feature_types=feature_types)
        np.testing.assert_equal(np.array(Xy.feature_types), np.array(feature_types))

    def test_uri_categorical(self):
        path = os.path.join(dpath, "agaricus.txt.train")
        feature_types = ["q"] * 5 + ["c"] + ["q"] * 120
        Xy = xgb.DMatrix(
            path + "?indexing_mode=1&format=libsvm", feature_types=feature_types
        )
        np.testing.assert_equal(np.array(Xy.feature_types), np.array(feature_types))

    def test_base_margin(self) -> None:
        run_base_margin_info(np.asarray, xgb.DMatrix, "cpu")

    @given(
        strategies.integers(0, 1000),
        strategies.integers(0, 100),
        strategies.fractions(0, 1),
    )
    @settings(deadline=None, print_blob=True)
    def test_to_csr(self, n_samples, n_features, sparsity) -> None:
        if n_samples == 0 or n_features == 0 or sparsity == 1.0:
            csr = scipy.sparse.csr_matrix(np.empty((0, 0)))
        else:
            csr = tm.make_sparse_regression(n_samples, n_features, sparsity, False)[
                0
            ].astype(np.float32)
        m = xgb.DMatrix(data=csr)
        ret = m.get_data()
        np.testing.assert_equal(csr.indptr, ret.indptr)
        np.testing.assert_equal(csr.data, ret.data)
        np.testing.assert_equal(csr.indices, ret.indices)

    def test_dtypes(self) -> None:
        n_samples = 128
        n_features = 16
        for orig, x in np_dtypes(n_samples, n_features):
            m0 = xgb.DMatrix(orig)
            m1 = xgb.DMatrix(x)
            assert predictor_equal(m0, m1)


@pytest.mark.skipif(tm.is_windows(), reason="Rabit does not run on windows")
class TestDMatrixColumnSplit:
    def test_numpy(self):
        def verify_numpy():
            data = np.random.randn(5, 5)
            dm = xgb.DMatrix(data, data_split_mode=DataSplitMode.COL)
            assert dm.num_row() == 5
            assert dm.num_col() == 5 * xgb.collective.get_world_size()
            assert dm.feature_names is None
            assert dm.feature_types is None

        tm.run_with_rabit(world_size=3, test_fn=verify_numpy)

    def test_numpy_feature_names(self):
        def verify_numpy_feature_names():
            world_size = xgb.collective.get_world_size()
            data = np.random.randn(5, 5)
            feature_names = [f"feature{x}" for x in range(5)]
            feature_types = ["float"] * 5
            dm = xgb.DMatrix(
                data,
                feature_names=feature_names,
                feature_types=feature_types,
                data_split_mode=DataSplitMode.COL,
            )
            assert dm.num_row() == 5
            assert dm.num_col() == 5 * world_size
            assert len(dm.feature_names) == 5 * world_size
            assert dm.feature_names == tm.column_split_feature_names(
                feature_names, world_size
            )
            assert len(dm.feature_types) == 5 * world_size
            assert dm.feature_types == ["float"] * 5 * world_size

        tm.run_with_rabit(world_size=3, test_fn=verify_numpy_feature_names)

    def test_csr(self):
        def verify_csr():
            indptr = np.array([0, 2, 3, 6])
            indices = np.array([0, 2, 2, 0, 1, 2])
            data = np.array([1, 2, 3, 4, 5, 6])
            X = scipy.sparse.csr_matrix((data, indices, indptr), shape=(3, 3))
            dtrain = xgb.DMatrix(X, data_split_mode=DataSplitMode.COL)
            assert dtrain.num_row() == 3
            assert dtrain.num_col() == 3 * xgb.collective.get_world_size()

        tm.run_with_rabit(world_size=3, test_fn=verify_csr)

    def test_csc(self):
        def verify_csc():
            row = np.array([0, 2, 2, 0, 1, 2])
            col = np.array([0, 0, 1, 2, 2, 2])
            data = np.array([1, 2, 3, 4, 5, 6])
            X = scipy.sparse.csc_matrix((data, (row, col)), shape=(3, 3))
            dtrain = xgb.DMatrix(X, data_split_mode=DataSplitMode.COL)
            assert dtrain.num_row() == 3
            assert dtrain.num_col() == 3 * xgb.collective.get_world_size()

        tm.run_with_rabit(world_size=3, test_fn=verify_csc)

    def test_coo(self):
        def verify_coo():
            row = np.array([0, 2, 2, 0, 1, 2])
            col = np.array([0, 0, 1, 2, 2, 2])
            data = np.array([1, 2, 3, 4, 5, 6])
            X = scipy.sparse.coo_matrix((data, (row, col)), shape=(3, 3))
            dtrain = xgb.DMatrix(X, data_split_mode=DataSplitMode.COL)
            assert dtrain.num_row() == 3
            assert dtrain.num_col() == 3 * xgb.collective.get_world_size()

        tm.run_with_rabit(world_size=3, test_fn=verify_coo)

    def test_uri(self, tmp_path: Path) -> None:
        def verify_uri():
            rank = xgb.collective.get_rank()
            filename = tmp_path / f"test_data_{rank}.csv"

            data = np.random.rand(5, 5)
            with open(filename, mode="w", newline="") as file:
                writer = csv.writer(file)
                for row in data:
                    writer.writerow(row)
            dtrain = xgb.DMatrix(
                f"{filename}?format=csv", data_split_mode=DataSplitMode.COL
            )
            assert dtrain.num_row() == 5
            assert dtrain.num_col() == 5 * xgb.collective.get_world_size()

        tm.run_with_rabit(world_size=3, test_fn=verify_uri)

    def test_list(self):
        def verify_list():
            data = [
                [1, 2, 3, 4, 5],
                [6, 7, 8, 9, 10],
                [11, 12, 13, 14, 15],
                [16, 17, 18, 19, 20],
                [21, 22, 23, 24, 25],
            ]
            dm = xgb.DMatrix(data, data_split_mode=DataSplitMode.COL)
            assert dm.num_row() == 5
            assert dm.num_col() == 5 * xgb.collective.get_world_size()

        tm.run_with_rabit(world_size=3, test_fn=verify_list)

    def test_tuple(self):
        def verify_tuple():
            data = (
                (1, 2, 3, 4, 5),
                (6, 7, 8, 9, 10),
                (11, 12, 13, 14, 15),
                (16, 17, 18, 19, 20),
                (21, 22, 23, 24, 25),
            )
            dm = xgb.DMatrix(data, data_split_mode=DataSplitMode.COL)
            assert dm.num_row() == 5
            assert dm.num_col() == 5 * xgb.collective.get_world_size()

        tm.run_with_rabit(world_size=3, test_fn=verify_tuple)



import json
import os
import pickle
import re
import warnings
from pathlib import Path
from typing import Optional

import numpy as np
import pytest
import xgboost as xgb
from sklearn.utils.estimator_checks import parametrize_with_checks
from xgboost import testing as tm
from xgboost.testing.data import get_california_housing
from xgboost.testing.ranking import run_ranking_categorical, run_ranking_qid_df
from xgboost.testing.shared import get_feature_weights, validate_data_initialization
from xgboost.testing.updater import get_basescore
from xgboost.testing.with_skl import (
    run_boost_from_prediction_binary,
    run_boost_from_prediction_multi_clasas,
    run_housing_rf_regression,
    run_intercept,
    run_recoding,
)

rng = np.random.RandomState(1994)
pytestmark = [pytest.mark.skipif(**tm.no_sklearn()), tm.timeout(30)]


def test_binary_classification():
    from sklearn.datasets import load_digits
    from sklearn.model_selection import KFold

    digits = load_digits(n_class=2)
    y = digits["target"]
    X = digits["data"]
    kf = KFold(n_splits=2, shuffle=True, random_state=rng)
    for cls in (xgb.XGBClassifier, xgb.XGBRFClassifier):
        for train_index, test_index in kf.split(X, y):
            clf = cls(random_state=42, eval_metric=["auc", "logloss"])
            xgb_model = clf.fit(X[train_index], y[train_index])
            preds = xgb_model.predict(X[test_index])
            labels = y[test_index]
            err = sum(
                1 for i in range(len(preds)) if int(preds[i] > 0.5) != labels[i]
            ) / float(len(preds))
            assert err < 0.1


@pytest.mark.parametrize("objective", ["multi:softmax", "multi:softprob"])
def test_multiclass_classification(objective):
    from sklearn.datasets import load_iris
    from sklearn.model_selection import KFold

    def check_pred(preds, labels, output_margin):
        if output_margin:
            err = sum(
                1 for i in range(len(preds)) if preds[i].argmax() != labels[i]
            ) / float(len(preds))
        else:
            err = sum(1 for i in range(len(preds)) if preds[i] != labels[i]) / float(
                len(preds)
            )
        assert err < 0.4

    X, y = load_iris(return_X_y=True)
    kf = KFold(n_splits=2, shuffle=True, random_state=rng)
    for train_index, test_index in kf.split(X, y):
        xgb_model = xgb.XGBClassifier(objective=objective).fit(
            X[train_index], y[train_index]
        )
        assert xgb_model.get_booster().num_boosted_rounds() == 100
        preds = xgb_model.predict(X[test_index])
        # test other params in XGBClassifier().fit
        preds2 = xgb_model.predict(
            X[test_index], output_margin=True, iteration_range=(0, 1)
        )
        preds3 = xgb_model.predict(
            X[test_index], output_margin=True, iteration_range=None
        )
        preds4 = xgb_model.predict(
            X[test_index], output_margin=False, iteration_range=(0, 1)
        )
        labels = y[test_index]

        check_pred(preds, labels, output_margin=False)
        check_pred(preds2, labels, output_margin=True)
        check_pred(preds3, labels, output_margin=True)
        check_pred(preds4, labels, output_margin=False)

    cls = xgb.XGBClassifier(n_estimators=4).fit(X, y)
    assert cls.n_classes_ == 3
    proba = cls.predict_proba(X)
    assert proba.shape[0] == X.shape[0]
    assert proba.shape[1] == cls.n_classes_

    # custom objective, the default is multi:softprob so no transformation is required.
    cls = xgb.XGBClassifier(n_estimators=4, objective=tm.softprob_obj(3)).fit(X, y)
    proba = cls.predict_proba(X)
    assert proba.shape[0] == X.shape[0]
    assert proba.shape[1] == cls.n_classes_


def test_best_iteration():
    from sklearn.datasets import load_iris

    X, y = load_iris(return_X_y=True)

    def train(booster: str, forest: Optional[int]) -> None:
        rounds = 4
        cls = xgb.XGBClassifier(
            n_estimators=rounds,
            num_parallel_tree=forest,
            booster=booster,
            early_stopping_rounds=3,
        ).fit(X, y, eval_set=[(X, y)])
        assert cls.best_iteration == rounds - 1

        # best_iteration is used by default, assert that under gblinear it's
        # automatically ignored due to being 0.
        cls.predict(X)

    num_parallel_tree = 4
    train("gbtree", num_parallel_tree)
    train("dart", num_parallel_tree)
    train("gblinear", None)


def test_ranking():
    # generate random data
    x_train = np.random.rand(1000, 10)
    y_train = np.random.randint(5, size=1000)
    train_group = np.repeat(50, 20)

    x_valid = np.random.rand(200, 10)
    y_valid = np.random.randint(5, size=200)
    valid_group = np.repeat(50, 4)

    x_test = np.random.rand(100, 10)

    params = {
        "tree_method": "exact",
        "objective": "rank:pairwise",
        "learning_rate": 0.1,
        "gamma": 1.0,
        "min_child_weight": 0.1,
        "max_depth": 6,
        "n_estimators": 4,
    }
    model = xgb.sklearn.XGBRanker(**params)
    model.fit(
        x_train,
        y_train,
        group=train_group,
        eval_set=[(x_valid, y_valid)],
        eval_group=[valid_group],
    )
    assert model.evals_result()

    pred = model.predict(x_test)

    train_data = xgb.DMatrix(x_train, y_train)
    valid_data = xgb.DMatrix(x_valid, y_valid)
    test_data = xgb.DMatrix(x_test)
    train_data.set_group(train_group)
    assert train_data.get_label().shape[0] == x_train.shape[0]
    valid_data.set_group(valid_group)

    params_orig = {
        "tree_method": "exact",
        "objective": "rank:pairwise",
        "eta": 0.1,
        "gamma": 1.0,
        "min_child_weight": 0.1,
        "max_depth": 6,
    }
    xgb_model_orig = xgb.train(
        params_orig, train_data, num_boost_round=4, evals=[(valid_data, "validation")]
    )
    pred_orig = xgb_model_orig.predict(test_data)

    np.testing.assert_almost_equal(pred, pred_orig)


@pytest.mark.skipif(**tm.no_pandas())
def test_ranking_categorical() -> None:
    run_ranking_categorical(device="cpu")


def test_ranking_metric() -> None:
    from sklearn.metrics import roc_auc_score

    X, y, qid, w = tm.make_ltr(512, 4, 3, 1)
    # use auc for test as ndcg_score in sklearn works only on label gain instead of exp
    # gain.
    # note that the auc in sklearn is different from the one in XGBoost. The one in
    # sklearn compares the number of mis-classified docs, while the one in xgboost
    # compares the number of mis-classified pairs.
    ltr = xgb.XGBRanker(
        eval_metric=roc_auc_score,
        n_estimators=10,
        tree_method="hist",
        max_depth=2,
        objective="rank:pairwise",
    )
    ltr.fit(
        X,
        y,
        qid=qid,
        sample_weight=w,
        eval_set=[(X, y)],
        eval_qid=[qid],
        sample_weight_eval_set=[w],
        verbose=True,
    )
    results = ltr.evals_result()
    assert results["validation_0"]["roc_auc_score"][-1] > 0.6


@pytest.mark.skipif(**tm.no_pandas())
def test_ranking_qid_df():
    import pandas as pd

    run_ranking_qid_df(pd, "hist", "cpu")


def test_stacking_regression():
    from sklearn.datasets import load_diabetes
    from sklearn.ensemble import RandomForestRegressor, StackingRegressor
    from sklearn.linear_model import RidgeCV
    from sklearn.model_selection import train_test_split

    X, y = load_diabetes(return_X_y=True)
    estimators = [
        ("gbm", xgb.sklearn.XGBRegressor(objective="reg:squarederror")),
        ("lr", RidgeCV()),
    ]
    reg = StackingRegressor(
        estimators=estimators,
        final_estimator=RandomForestRegressor(n_estimators=10, random_state=42),
    )

    X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=42)
    reg.fit(X_train, y_train).score(X_test, y_test)


def test_stacking_classification():
    from sklearn.datasets import load_iris
    from sklearn.ensemble import StackingClassifier
    from sklearn.linear_model import LogisticRegression
    from sklearn.model_selection import train_test_split
    from sklearn.pipeline import make_pipeline
    from sklearn.preprocessing import StandardScaler
    from sklearn.svm import LinearSVC

    X, y = load_iris(return_X_y=True)
    estimators = [
        ("gbm", xgb.sklearn.XGBClassifier()),
        ("svr", make_pipeline(StandardScaler(), LinearSVC(random_state=42))),
    ]
    clf = StackingClassifier(
        estimators=estimators, final_estimator=LogisticRegression()
    )

    X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=42)
    clf.fit(X_train, y_train).score(X_test, y_test)


@pytest.mark.skipif(**tm.no_pandas())
def test_feature_importances_weight(tmp_path: Path) -> None:
    from sklearn.datasets import load_digits

    digits = load_digits(n_class=2)
    y = digits["target"]
    X = digits["data"]

    xgb_model = xgb.XGBClassifier(
        random_state=0,
        tree_method="exact",
        learning_rate=0.1,
        importance_type="weight",
        base_score=0.5,
    ).fit(X, y)
    # fmt: off
    exp = np.array([0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.00833333, 0.,
                    0., 0., 0., 0., 0., 0., 0., 0.025, 0.14166667, 0., 0., 0.,
                    0., 0., 0., 0.00833333, 0.25833333, 0., 0., 0., 0.,
                    0.03333334, 0.03333334, 0., 0.32499999, 0., 0., 0., 0.,
                    0.05, 0.06666667, 0., 0., 0., 0., 0., 0., 0., 0.04166667,
                    0., 0., 0., 0., 0., 0., 0., 0.00833333, 0., 0., 0., 0.,
                    0.], dtype=np.float32)
    # fmt: on
    np.testing.assert_almost_equal(xgb_model.feature_importances_, exp)

    # numeric columns
    import pandas as pd

    y = pd.Series(digits["target"])
    X = pd.DataFrame(digits["data"])
    xgb_model = xgb.XGBClassifier(
        random_state=0,
        tree_method="exact",
        learning_rate=0.1,
        base_score=0.5,
        importance_type="weight",
    ).fit(X, y)
    np.testing.assert_almost_equal(xgb_model.feature_importances_, exp)

    xgb_model = xgb.XGBClassifier(
        random_state=0,
        tree_method="exact",
        learning_rate=0.1,
        importance_type="weight",
        base_score=0.5,
    ).fit(X, y)
    np.testing.assert_almost_equal(xgb_model.feature_importances_, exp)

    with pytest.raises(ValueError):
        xgb_model.set_params(importance_type="foo")
        xgb_model.feature_importances_

    X, y = load_digits(n_class=3, return_X_y=True)

    cls = xgb.XGBClassifier(booster="gblinear", n_estimators=4)
    cls.fit(X, y)
    assert cls.feature_importances_.shape[0] == X.shape[1]
    assert cls.feature_importances_.shape[1] == 3
    path = tmp_path / "model.json"
    cls.save_model(path)
    with open(path, "r") as fd:
        model = json.load(fd)
    weights = np.array(
        model["learner"]["gradient_booster"]["model"]["weights"]
    ).reshape((cls.n_features_in_ + 1, 3))
    weights = weights[:-1, ...]
    np.testing.assert_allclose(
        weights / weights.sum(), cls.feature_importances_, rtol=1e-6
    )

    with pytest.raises(ValueError):
        cls.set_params(importance_type="cover")
        cls.feature_importances_


@pytest.mark.skipif(**tm.no_pandas())
def test_feature_importances_gain():
    from sklearn.datasets import load_digits

    digits = load_digits(n_class=2)
    y = digits["target"]
    X = digits["data"]
    xgb_model = xgb.XGBClassifier(
        random_state=0,
        tree_method="exact",
        learning_rate=0.1,
        importance_type="gain",
        base_score=0.5,
    ).fit(X, y)
    # fmt: off
    exp = np.array([0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
                    0.00326159, 0., 0., 0., 0., 0., 0., 0., 0.,
                    0.00297238, 0.00988034, 0., 0., 0., 0., 0., 0.,
                    0.03512521, 0.41123885, 0., 0., 0., 0.,
                    0.01326332, 0.00160674, 0., 0.4206952, 0., 0., 0.,
                    0., 0.00616747, 0.01237546, 0., 0., 0., 0., 0.,
                    0., 0., 0.08240705, 0., 0., 0., 0., 0., 0., 0.,
                    0.00100649, 0., 0., 0., 0., 0.], dtype=np.float32)
    # fmt: on
    np.testing.assert_almost_equal(xgb_model.feature_importances_, exp)

    # numeric columns
    import pandas as pd

    y = pd.Series(digits["target"])
    X = pd.DataFrame(digits["data"])
    xgb_model = xgb.XGBClassifier(
        random_state=0,
        tree_method="exact",
        learning_rate=0.1,
        importance_type="gain",
        base_score=0.5,
    ).fit(X, y)
    np.testing.assert_almost_equal(xgb_model.feature_importances_, exp)

    xgb_model = xgb.XGBClassifier(
        random_state=0,
        tree_method="exact",
        learning_rate=0.1,
        importance_type="gain",
        base_score=0.5,
    ).fit(X, y)
    np.testing.assert_almost_equal(xgb_model.feature_importances_, exp)

    # no split can be found
    cls = xgb.XGBClassifier(min_child_weight=1000, tree_method="hist", n_estimators=1)
    cls.fit(X, y)
    assert np.all(cls.feature_importances_ == 0)


def test_select_feature():
    from sklearn.datasets import load_digits
    from sklearn.feature_selection import SelectFromModel

    digits = load_digits(n_class=2)
    y = digits["target"]
    X = digits["data"]
    cls = xgb.XGBClassifier()
    cls.fit(X, y)
    selector = SelectFromModel(cls, prefit=True, max_features=1)
    X_selected = selector.transform(X)
    assert X_selected.shape[1] == 1


def test_num_parallel_tree():
    from sklearn.datasets import load_diabetes

    reg = xgb.XGBRegressor(n_estimators=4, num_parallel_tree=4, tree_method="hist")
    X, y = load_diabetes(return_X_y=True)
    bst = reg.fit(X=X, y=y)
    dump = bst.get_booster().get_dump(dump_format="json")
    assert len(dump) == 16

    reg = xgb.XGBRFRegressor(n_estimators=4)
    bst = reg.fit(X=X, y=y)
    dump = bst.get_booster().get_dump(dump_format="json")
    assert len(dump) == 4

    config = json.loads(bst.get_booster().save_config())
    assert (
        int(
            config["learner"]["gradient_booster"]["gbtree_model_param"][
                "num_parallel_tree"
            ]
        )
        == 4
    )


def test_regression():
    from sklearn.metrics import mean_squared_error
    from sklearn.model_selection import KFold

    X, y = get_california_housing()
    kf = KFold(n_splits=2, shuffle=True, random_state=rng)
    for train_index, test_index in kf.split(X, y):
        xgb_model = xgb.XGBRegressor().fit(X[train_index], y[train_index])

        preds = xgb_model.predict(X[test_index])
        # test other params in XGBRegressor().fit
        preds2 = xgb_model.predict(
            X[test_index], output_margin=True, iteration_range=(0, np.int16(3))
        )
        preds3 = xgb_model.predict(
            X[test_index], output_margin=True, iteration_range=None
        )
        preds4 = xgb_model.predict(
            X[test_index], output_margin=False, iteration_range=(0, 3)
        )
        labels = y[test_index]

        assert mean_squared_error(preds, labels) < 25
        assert mean_squared_error(preds2, labels) < 350
        assert mean_squared_error(preds3, labels) < 25
        assert mean_squared_error(preds4, labels) < 350

        with pytest.raises(AttributeError, match="feature_names_in_"):
            xgb_model.feature_names_in_


def test_rf_regression():
    run_housing_rf_regression("hist", "cpu")


@pytest.mark.parametrize("tree_method", ["exact", "hist", "approx"])
def test_parameter_tuning(tree_method: str) -> None:
    from sklearn.model_selection import GridSearchCV

    X, y = get_california_housing()
    reg = xgb.XGBRegressor(learning_rate=0.1, tree_method=tree_method)
    grid_cv = GridSearchCV(
        reg, {"max_depth": [2, 4], "n_estimators": [50, 200]}, cv=2, verbose=1
    )
    grid_cv.fit(X, y)
    assert grid_cv.best_score_ < 0.7
    assert grid_cv.best_params_ == {
        "n_estimators": 50,
        "max_depth": 2,
    }


def test_regression_with_custom_objective():
    from sklearn.metrics import mean_squared_error
    from sklearn.model_selection import KFold

    X, y = get_california_housing()
    kf = KFold(n_splits=2, shuffle=True, random_state=rng)
    for train_index, test_index in kf.split(X, y):
        xgb_model = xgb.XGBRegressor(objective=tm.ls_obj).fit(
            X[train_index], y[train_index]
        )
        preds = xgb_model.predict(X[test_index])
        labels = y[test_index]
    assert mean_squared_error(preds, labels) < 25

    w = rng.uniform(low=0.0, high=1.0, size=X.shape[0])
    reg = xgb.XGBRegressor(objective=tm.ls_obj, n_estimators=25)
    reg.fit(X, y, sample_weight=w)
    y_pred = reg.predict(X)
    assert mean_squared_error(y_true=y, y_pred=y_pred, sample_weight=w) < 25

    # Test that the custom objective function is actually used
    class XGBCustomObjectiveException(Exception):
        pass

    def dummy_objective(y_true, y_pred):
        raise XGBCustomObjectiveException()

    xgb_model = xgb.XGBRegressor(objective=dummy_objective)
    np.testing.assert_raises(XGBCustomObjectiveException, xgb_model.fit, X, y)


def logregobj(y_true, y_pred):
    y_pred = 1.0 / (1.0 + np.exp(-y_pred))
    grad = y_pred - y_true
    hess = y_pred * (1.0 - y_pred)
    return grad, hess


def test_classification_with_custom_objective():
    from sklearn.datasets import load_digits
    from sklearn.model_selection import KFold

    digits = load_digits(n_class=2)
    y = digits["target"]
    X = digits["data"]
    kf = KFold(n_splits=2, shuffle=True, random_state=rng)
    for train_index, test_index in kf.split(X, y):
        xgb_model = xgb.XGBClassifier(objective=logregobj)
        xgb_model.fit(X[train_index], y[train_index])
        preds = xgb_model.predict(X[test_index])
        labels = y[test_index]
        err = sum(
            1 for i in range(len(preds)) if int(preds[i] > 0.5) != labels[i]
        ) / float(len(preds))
        assert err < 0.1

    # Test that the custom objective function is actually used
    class XGBCustomObjectiveException(Exception):
        pass

    def dummy_objective(y_true, y_preds):
        raise XGBCustomObjectiveException()

    xgb_model = xgb.XGBClassifier(objective=dummy_objective)
    np.testing.assert_raises(XGBCustomObjectiveException, xgb_model.fit, X, y)

    cls = xgb.XGBClassifier(n_estimators=1)
    cls.fit(X, y)

    is_called = [False]

    def wrapped(y, p):
        is_called[0] = True
        return logregobj(y, p)

    cls.set_params(objective=wrapped)
    cls.predict(X)  # no throw
    cls.fit(X, y)

    assert is_called[0]


def run_sklearn_api(booster, error, n_est):
    from sklearn.datasets import load_iris
    from sklearn.model_selection import train_test_split

    iris = load_iris()
    tr_d, te_d, tr_l, te_l = train_test_split(
        iris.data, iris.target, train_size=120, test_size=0.2
    )

    classifier = xgb.XGBClassifier(booster=booster, n_estimators=n_est)
    classifier.fit(tr_d, tr_l)

    preds = classifier.predict(te_d)
    labels = te_l
    err = sum([1 for p, l in zip(preds, labels) if p != l]) * 1.0 / len(te_l)
    assert err < error


def test_sklearn_api():
    run_sklearn_api("gbtree", 0.2, 10)
    run_sklearn_api("gblinear", 0.5, 100)


@pytest.mark.skipif(**tm.no_matplotlib())
@pytest.mark.skipif(**tm.no_graphviz())
def test_sklearn_plotting():
    from sklearn.datasets import load_iris

    iris = load_iris()

    classifier = xgb.XGBClassifier()
    classifier.fit(iris.data, iris.target)

    import matplotlib

    matplotlib.use("Agg")

    from graphviz import Source
    from matplotlib.axes import Axes

    ax = xgb.plot_importance(classifier)
    assert isinstance(ax, Axes)
    assert ax.get_title() == "Feature importance"
    assert ax.get_xlabel() == "Importance score"
    assert ax.get_ylabel() == "Features"
    assert len(ax.patches) == 4

    g = xgb.to_graphviz(classifier, num_trees=0)
    assert isinstance(g, Source)

    ax = xgb.plot_tree(classifier, num_trees=0)
    assert isinstance(ax, Axes)


@pytest.mark.skipif(**tm.no_pandas())
def test_sklearn_nfolds_cv():
    from sklearn.datasets import load_digits
    from sklearn.model_selection import StratifiedKFold

    digits = load_digits(n_class=3)
    X = digits["data"]
    y = digits["target"]
    dm = xgb.DMatrix(X, label=y)

    params = {
        "max_depth": 2,
        "eta": 1,
        "verbosity": 0,
        "objective": "multi:softprob",
        "num_class": 3,
    }

    seed = 2016
    nfolds = 5
    skf = StratifiedKFold(n_splits=nfolds, shuffle=True, random_state=seed)

    cv1 = xgb.cv(
        params, dm, num_boost_round=10, nfold=nfolds, seed=seed, as_pandas=True
    )
    cv2 = xgb.cv(
        params,
        dm,
        num_boost_round=10,
        nfold=nfolds,
        folds=skf,
        seed=seed,
        as_pandas=True,
    )
    cv3 = xgb.cv(
        params,
        dm,
        num_boost_round=10,
        nfold=nfolds,
        stratified=True,
        seed=seed,
        as_pandas=True,
    )
    assert cv1.shape[0] == cv2.shape[0] and cv2.shape[0] == cv3.shape[0]
    assert cv2.iloc[-1, 0] == cv3.iloc[-1, 0]


@pytest.mark.skipif(**tm.no_pandas())
def test_split_value_histograms():
    from sklearn.datasets import load_digits

    digits_2class = load_digits(n_class=2)

    X = digits_2class["data"]
    y = digits_2class["target"]

    dm = xgb.DMatrix(X, label=y)
    params = {
        "max_depth": 6,
        "eta": 0.01,
        "objective": "binary:logistic",
        "base_score": 0.5,
    }

    gbdt = xgb.train(params, dm, num_boost_round=10)
    assert gbdt.get_split_value_histogram("not_there", as_pandas=True).shape[0] == 0
    assert gbdt.get_split_value_histogram("not_there", as_pandas=False).shape[0] == 0
    assert gbdt.get_split_value_histogram("f28", bins=0).shape[0] == 1
    assert gbdt.get_split_value_histogram("f28", bins=1).shape[0] == 1
    assert gbdt.get_split_value_histogram("f28", bins=2).shape[0] == 2
    assert gbdt.get_split_value_histogram("f28", bins=5).shape[0] == 2
    assert gbdt.get_split_value_histogram("f28", bins=None).shape[0] == 2


def test_sklearn_random_state():
    clf = xgb.XGBClassifier(random_state=402)
    assert clf.get_xgb_params()["random_state"] == 402

    clf = xgb.XGBClassifier(random_state=401)
    assert clf.get_xgb_params()["random_state"] == 401

    random_state = np.random.RandomState(seed=403)
    clf = xgb.XGBClassifier(random_state=random_state)
    assert isinstance(clf.get_xgb_params()["random_state"], int)

    random_state = np.random.default_rng(seed=404)
    clf = xgb.XGBClassifier(random_state=random_state)
    assert isinstance(clf.get_xgb_params()["random_state"], int)


def test_sklearn_n_jobs():
    clf = xgb.XGBClassifier(n_jobs=1)
    assert clf.get_xgb_params()["n_jobs"] == 1

    clf = xgb.XGBClassifier(n_jobs=2)
    assert clf.get_xgb_params()["n_jobs"] == 2


def test_parameters_access(tmp_path: Path) -> None:
    from sklearn import datasets

    params = {"updater": "grow_gpu_hist", "subsample": 0.5, "n_jobs": -1}
    clf = xgb.XGBClassifier(n_estimators=1000, **params)
    assert clf.get_params()["updater"] == "grow_gpu_hist"
    assert clf.get_params()["subsample"] == 0.5
    assert clf.get_params()["n_estimators"] == 1000

    clf = xgb.XGBClassifier(n_estimators=1, nthread=4)
    X, y = datasets.load_iris(return_X_y=True)
    clf.fit(X, y)

    config = json.loads(clf.get_booster().save_config())
    assert int(config["learner"]["generic_param"]["nthread"]) == 4

    clf.set_params(nthread=16)
    config = json.loads(clf.get_booster().save_config())
    assert int(config["learner"]["generic_param"]["nthread"]) == 16

    clf.predict(X)
    config = json.loads(clf.get_booster().save_config())
    assert int(config["learner"]["generic_param"]["nthread"]) == 16

    clf = xgb.XGBClassifier(n_estimators=2)
    assert clf.tree_method is None
    assert clf.get_params()["tree_method"] is None
    clf.fit(X, y)
    assert clf.get_params()["tree_method"] is None

    save_load_counter = [0]

    def save_load(clf: xgb.XGBClassifier) -> xgb.XGBClassifier:
        path = tmp_path / f"model_{save_load_counter[0]}.json"
        save_load_counter[0] += 1
        clf.save_model(path)
        clf = xgb.XGBClassifier()
        clf.load_model(path)
        return clf

    def get_tm(clf: xgb.XGBClassifier) -> str:
        tm_val = json.loads(clf.get_booster().save_config())["learner"][
            "gradient_booster"
        ]["gbtree_train_param"]["tree_method"]
        return tm_val

    assert get_tm(clf) == "auto"  # Kept as auto, immutable since 2.0

    clf = pickle.loads(pickle.dumps(clf))

    assert clf.tree_method is None
    assert clf.n_estimators == 2
    assert clf.get_params()["tree_method"] is None
    assert clf.get_params()["n_estimators"] == 2
    assert get_tm(clf) == "auto"  # preserved for pickle

    clf = save_load(clf)

    assert clf.tree_method is None
    assert clf.n_estimators is None
    assert clf.get_params()["tree_method"] is None
    assert clf.get_params()["n_estimators"] is None
    assert get_tm(clf) == "auto"  # discarded for save/load_model

    clf.set_params(tree_method="hist")
    assert clf.get_params()["tree_method"] == "hist"
    clf = pickle.loads(pickle.dumps(clf))
    assert clf.get_params()["tree_method"] == "hist"
    clf = save_load(clf)
    assert clf.get_params()["tree_method"] is None


def test_get_params_works_as_expected():
    # XGBModel -> BaseEstimator
    params = xgb.XGBModel(max_depth=2).get_params()
    assert params["max_depth"] == 2
    # 'objective' defaults to None in the signature of XGBModel
    assert params["objective"] is None

    # XGBRegressor -> XGBModel -> BaseEstimator
    params = xgb.XGBRegressor(max_depth=3).get_params()
    assert params["max_depth"] == 3
    # 'objective' defaults to 'reg:squarederror' in the signature of XGBRegressor
    assert params["objective"] == "reg:squarederror"
    # 'colsample_bynode' defaults to 'None' for XGBModel (which XGBRegressor inherits from), so it
    # should be in get_params() output
    assert params["colsample_bynode"] is None

    # XGBRFRegressor -> XGBRegressor -> XGBModel -> BaseEstimator
    params = xgb.XGBRFRegressor(max_depth=4, objective="reg:tweedie").get_params()
    assert params["max_depth"] == 4
    # 'objective' is a keyword argument for XGBRegressor, so it should be in get_params() output
    # ... but values passed through kwargs should override the default from the signature of XGBRegressor
    assert params["objective"] == "reg:tweedie"
    # 'colsample_bynode' defaults to 0.8 for XGBRFRegressor...that should be preferred to the None from XGBRegressor
    assert params["colsample_bynode"] == 0.8


def test_kwargs_error():
    params = {"updater": "grow_gpu_hist", "subsample": 0.5, "n_jobs": -1}
    with pytest.raises(TypeError):
        clf = xgb.XGBClassifier(n_jobs=1000, **params)
        assert isinstance(clf, xgb.XGBClassifier)


def test_kwargs_grid_search():
    from sklearn import datasets
    from sklearn.model_selection import GridSearchCV

    params = {"tree_method": "hist"}
    clf = xgb.XGBClassifier(n_estimators=3, **params)
    assert clf.get_params()["tree_method"] == "hist"
    # 'eta' is not a default argument of XGBClassifier
    # Check we can still do grid search over this parameter
    search_params = {"eta": [0, 0.2, 0.4]}
    grid_cv = GridSearchCV(clf, search_params, cv=5)
    iris = datasets.load_iris()
    grid_cv.fit(iris.data, iris.target)

    # Expect unique results for each parameter value
    # This confirms sklearn is able to successfully update the parameter
    means = grid_cv.cv_results_["mean_test_score"]
    assert len(means) == len(set(means))


def test_sklearn_clone():
    from sklearn.base import clone

    clf = xgb.XGBClassifier(n_jobs=2)
    clf.n_jobs = -1
    clone(clf)


def test_sklearn_get_default_params():
    from sklearn.datasets import load_digits

    digits_2class = load_digits(n_class=2)
    X = digits_2class["data"]
    y = digits_2class["target"]
    cls = xgb.XGBClassifier()
    assert cls.get_params()["base_score"] is None
    cls.fit(X[:4, ...], y[:4, ...])
    base_score = get_basescore(cls)
    np.testing.assert_equal(base_score, [0.5])


def run_validation_weights(model):
    from sklearn.datasets import make_hastie_10_2

    # prepare training and test data
    X, y = make_hastie_10_2(n_samples=2000, random_state=42)
    labels, y = np.unique(y, return_inverse=True)
    X_train, X_test = X[:1600], X[1600:]
    y_train, y_test = y[:1600], y[1600:]

    # instantiate model
    param_dist = {
        "objective": "binary:logistic",
        "n_estimators": 2,
        "random_state": 123,
    }
    clf = model(**param_dist)

    # train it using instance weights only in the training set
    weights_train = np.random.choice([1, 2], len(X_train))
    clf.set_params(eval_metric="logloss")
    clf.fit(
        X_train,
        y_train,
        sample_weight=weights_train,
        eval_set=[(X_test, y_test)],
        verbose=False,
    )
    # evaluate logloss metric on test set *without* using weights
    evals_result_without_weights = clf.evals_result()
    logloss_without_weights = evals_result_without_weights["validation_0"]["logloss"]

    # now use weights for the test set
    np.random.seed(0)
    weights_test = np.random.choice([1, 2], len(X_test))
    clf.set_params(eval_metric="logloss")
    clf.fit(
        X_train,
        y_train,
        sample_weight=weights_train,
        eval_set=[(X_test, y_test)],
        sample_weight_eval_set=[weights_test],
        verbose=False,
    )
    evals_result_with_weights = clf.evals_result()
    logloss_with_weights = evals_result_with_weights["validation_0"]["logloss"]

    # check that the logloss in the test set is actually different when using
    # weights than when not using them
    assert all((logloss_with_weights[i] != logloss_without_weights[i] for i in [0, 1]))

    with pytest.raises(ValueError):
        # length of eval set and sample weight doesn't match.
        clf.fit(
            X_train,
            y_train,
            sample_weight=weights_train,
            eval_set=[(X_train, y_train), (X_test, y_test)],
            sample_weight_eval_set=[weights_train],
        )

    with pytest.raises(ValueError):
        cls = xgb.XGBClassifier()
        cls.fit(
            X_train,
            y_train,
            sample_weight=weights_train,
            eval_set=[(X_train, y_train), (X_test, y_test)],
            sample_weight_eval_set=[weights_train],
        )


def test_validation_weights():
    run_validation_weights(xgb.XGBModel)
    run_validation_weights(xgb.XGBClassifier)


def test_RFECV():
    from sklearn.datasets import load_breast_cancer, load_diabetes, load_iris
    from sklearn.feature_selection import RFECV

    # Regression
    X, y = load_diabetes(return_X_y=True)
    bst = xgb.XGBRegressor(
        booster="gblinear",
        learning_rate=0.1,
        n_estimators=10,
        objective="reg:squarederror",
        random_state=0,
        verbosity=0,
    )
    rfecv = RFECV(estimator=bst, step=1, cv=3, scoring="neg_mean_squared_error")
    rfecv.fit(X, y)

    # Binary classification
    X, y = load_breast_cancer(return_X_y=True)
    bst = xgb.XGBClassifier(
        booster="gblinear",
        learning_rate=0.1,
        n_estimators=10,
        objective="binary:logistic",
        random_state=0,
        verbosity=0,
    )
    rfecv = RFECV(estimator=bst, step=0.5, cv=3, scoring="roc_auc")
    rfecv.fit(X, y)

    # Multi-class classification
    X, y = load_iris(return_X_y=True)
    bst = xgb.XGBClassifier(
        base_score=0.4,
        booster="gblinear",
        learning_rate=0.1,
        n_estimators=10,
        objective="multi:softprob",
        random_state=0,
        reg_alpha=0.001,
        reg_lambda=0.01,
        scale_pos_weight=0.5,
        verbosity=0,
    )
    rfecv = RFECV(estimator=bst, step=0.5, cv=3, scoring="neg_log_loss")
    rfecv.fit(X, y)

    X[0:4, :] = np.nan  # verify scikit_learn doesn't throw with nan
    reg = xgb.XGBRegressor()
    rfecv = RFECV(estimator=reg)
    rfecv.fit(X, y)

    cls = xgb.XGBClassifier()
    rfecv = RFECV(estimator=cls, step=0.5, cv=3, scoring="neg_mean_squared_error")
    rfecv.fit(X, y)


def test_XGBClassifier_resume(tmp_path: Path) -> None:
    from sklearn.datasets import load_breast_cancer
    from sklearn.metrics import log_loss

    model1_path = tmp_path / "test_XGBClassifier.model"
    model1_booster_path = tmp_path / "test_XGBClassifier.booster"

    X, Y = load_breast_cancer(return_X_y=True)

    model1 = xgb.XGBClassifier(learning_rate=0.3, random_state=0, n_estimators=8)
    model1.fit(X, Y)

    pred1 = model1.predict(X)
    log_loss1 = log_loss(pred1, Y)

    # file name of stored xgb model
    model1.save_model(model1_path)
    model2 = xgb.XGBClassifier(learning_rate=0.3, random_state=0, n_estimators=8)
    model2.fit(X, Y, xgb_model=model1_path)

    pred2 = model2.predict(X)
    log_loss2 = log_loss(pred2, Y)

    assert np.any(pred1 != pred2)
    assert log_loss1 > log_loss2

    # file name of 'Booster' instance Xgb model
    model1.get_booster().save_model(model1_booster_path)
    model2 = xgb.XGBClassifier(learning_rate=0.3, random_state=0, n_estimators=8)
    model2.fit(X, Y, xgb_model=model1_booster_path)

    pred2 = model2.predict(X)
    log_loss2 = log_loss(pred2, Y)

    assert np.any(pred1 != pred2)
    assert log_loss1 > log_loss2


def test_constraint_parameters():
    reg = xgb.XGBRegressor(interaction_constraints="[[0, 1], [2, 3, 4]]")
    X = np.random.randn(10, 10)
    y = np.random.randn(10)
    reg.fit(X, y)

    config = json.loads(reg.get_booster().save_config())
    assert (
        config["learner"]["gradient_booster"]["tree_train_param"][
            "interaction_constraints"
        ]
        == "[[0, 1], [2, 3, 4]]"
    )


@pytest.mark.filterwarnings("error")
def test_parameter_validation():
    reg = xgb.XGBRegressor(foo="bar", verbosity=1)
    X = np.random.randn(10, 10)
    y = np.random.randn(10)
    with pytest.warns(Warning, match="foo"):
        reg.fit(X, y)

    reg = xgb.XGBRegressor(
        n_estimators=2, missing=3, importance_type="gain", verbosity=1
    )
    X = np.random.randn(10, 10)
    y = np.random.randn(10)

    with warnings.catch_warnings():
        reg.fit(X, y)


def test_deprecate_position_arg():
    from sklearn.datasets import load_digits

    X, y = load_digits(return_X_y=True, n_class=2)
    w = np.random.default_rng(0).uniform(size=y.size)
    with pytest.warns(FutureWarning):
        xgb.XGBRegressor(3, learning_rate=0.1)
    model = xgb.XGBRegressor(n_estimators=1)
    with pytest.warns(FutureWarning):
        model.fit(X, y, w)

    with pytest.warns(FutureWarning):
        xgb.XGBClassifier(1)
    model = xgb.XGBClassifier(n_estimators=1)
    with pytest.warns(FutureWarning):
        model.fit(X, y, w)

    with pytest.warns(FutureWarning):
        xgb.XGBRanker("rank:ndcg", learning_rate=0.1)
    model = xgb.XGBRanker(n_estimators=1)
    group = np.repeat(1, X.shape[0])
    with pytest.warns(FutureWarning):
        model.fit(X, y, group)

    with pytest.warns(FutureWarning):
        xgb.XGBRFRegressor(1, learning_rate=0.1)
    model = xgb.XGBRFRegressor(n_estimators=1)
    with pytest.warns(FutureWarning):
        model.fit(X, y, w)

    model = xgb.XGBRFClassifier(n_estimators=1)
    with pytest.warns(FutureWarning):
        model.fit(X, y, w)


@pytest.mark.skipif(**tm.no_pandas())
def test_pandas_input():
    import pandas as pd
    from sklearn.calibration import CalibratedClassifierCV
    from sklearn.frozen import FrozenEstimator

    rng = np.random.RandomState(1994)

    kRows = 100
    kCols = 6

    X = rng.randint(low=0, high=2, size=kRows * kCols)
    X = X.reshape(kRows, kCols)

    df = pd.DataFrame(X)
    feature_names = []
    for i in range(1, kCols):
        feature_names += ["k" + str(i)]

    df.columns = ["status"] + feature_names

    target = df["status"]
    train = df.drop(columns=["status"])
    model = xgb.XGBClassifier()
    model.fit(train, target)
    np.testing.assert_equal(model.feature_names_in_, np.array(feature_names))

    columns = list(train.columns)
    rng.shuffle(columns)
    df_incorrect = df[columns]

    with pytest.raises(ValueError, match="feature_names mismatch"):
        model.predict(df_incorrect)

    clf_isotonic = CalibratedClassifierCV(FrozenEstimator(model), method="isotonic")
    clf_isotonic.fit(train, target)
    assert isinstance(
        clf_isotonic.calibrated_classifiers_[0].estimator.estimator, xgb.XGBClassifier
    )
    np.testing.assert_allclose(np.array(clf_isotonic.classes_), np.array([0, 1]))

    train_ser = train["k1"]
    assert isinstance(train_ser, pd.Series)
    model = xgb.XGBClassifier(n_estimators=8)
    model.fit(train_ser, target, eval_set=[(train_ser, target)])
    assert tm.non_increasing(model.evals_result()["validation_0"]["logloss"])


@pytest.mark.parametrize("tree_method", ["approx", "hist"])
def test_feature_weights(tree_method):
    kRows = 512
    kCols = 64
    X = rng.randn(kRows, kCols)
    y = rng.randn(kRows)

    fw = np.ones(shape=(kCols,))
    for i in range(kCols):
        fw[i] *= float(i)

    parser_path = os.path.join(tm.demo_dir(__file__), "guide-python", "model_parser.py")
    poly_increasing = get_feature_weights(
        X=X,
        y=y,
        fw=fw,
        parser_path=parser_path,
        tree_method=tree_method,
        model=xgb.XGBRegressor,
    )

    fw = np.ones(shape=(kCols,))
    for i in range(kCols):
        fw[i] *= float(kCols - i)
    poly_decreasing = get_feature_weights(
        X=X,
        y=y,
        fw=fw,
        parser_path=parser_path,
        tree_method=tree_method,
        model=xgb.XGBRegressor,
    )

    # Approxmated test, this is dependent on the implementation of random
    # number generator in std library.
    assert poly_increasing[0] > 0.08
    assert poly_decreasing[0] < -0.08

    reg = xgb.XGBRegressor(feature_weights=np.ones((kCols,)))
    with pytest.raises(ValueError, match="Use the one in"):
        reg.fit(X, y, feature_weights=np.ones((kCols,)))


@pytest.mark.parametrize("tree_method", ["hist", "approx", "exact"])
def test_boost_from_prediction(tree_method: str) -> None:
    import pandas as pd
    from sklearn.datasets import load_breast_cancer, load_iris, make_regression

    X, y = load_breast_cancer(return_X_y=True)

    run_boost_from_prediction_binary(tree_method, "cpu", X, y, None)
    run_boost_from_prediction_binary(tree_method, "cpu", X, y, pd.DataFrame)

    X, y = load_iris(return_X_y=True)

    run_boost_from_prediction_multi_clasas(
        xgb.XGBClassifier, tree_method, "cpu", X, y, None
    )
    run_boost_from_prediction_multi_clasas(
        xgb.XGBClassifier, tree_method, "cpu", X, y, pd.DataFrame
    )

    X, y = make_regression(n_samples=100, n_targets=4)
    run_boost_from_prediction_multi_clasas(
        xgb.XGBRegressor, tree_method, "cpu", X, y, None
    )


def test_estimator_type(tmp_path: Path) -> None:
    assert xgb.XGBClassifier()._get_type() == "classifier"
    assert xgb.XGBRFClassifier()._get_type() == "classifier"
    assert xgb.XGBRegressor()._get_type() == "regressor"
    assert xgb.XGBRFRegressor()._get_type() == "regressor"
    assert xgb.XGBRanker()._get_type() == "ranker"

    from sklearn.datasets import load_digits

    X, y = load_digits(n_class=2, return_X_y=True)
    cls = xgb.XGBClassifier(n_estimators=2).fit(X, y)
    path = tmp_path / "cls.json"
    cls.save_model(path)

    reg = xgb.XGBRegressor()
    with pytest.raises(TypeError):
        reg.load_model(path)

    cls = xgb.XGBClassifier()
    cls.load_model(path)  # no error


def test_multilabel_classification() -> None:
    from sklearn.datasets import make_multilabel_classification

    X, y = make_multilabel_classification(
        n_samples=32, n_classes=5, n_labels=3, random_state=0
    )
    clf = xgb.XGBClassifier(tree_method="hist")
    clf.fit(X, y)
    booster = clf.get_booster()
    learner = json.loads(booster.save_config())["learner"]
    assert int(learner["learner_model_param"]["num_target"]) == 5

    np.testing.assert_allclose(clf.predict(X), y)
    predt = (clf.predict_proba(X) > 0.5).astype(np.int64)
    np.testing.assert_allclose(clf.predict(X), predt)
    assert predt.dtype == np.int64

    y = y.tolist()
    clf.fit(X, y)
    np.testing.assert_allclose(clf.predict(X), predt)


def test_data_initialization() -> None:
    from sklearn.datasets import load_digits

    X, y = load_digits(return_X_y=True)
    validate_data_initialization(xgb.QuantileDMatrix, xgb.XGBClassifier, X, y)


@parametrize_with_checks([xgb.XGBRegressor()])
def test_estimator_reg(estimator, check):
    if os.environ["PYTEST_CURRENT_TEST"].find("check_supervised_y_no_nan") != -1:
        # The test uses float64 and requires the error message to contain:
        #
        #   "value too large for dtype(float64)",
        #
        # while XGBoost stores values as float32.  But XGBoost does verify the label
        # internally, so we replace this test with custom check.
        rng = np.random.RandomState(888)
        X = rng.randn(10, 5)
        y = np.full(10, np.inf)
        with pytest.raises(
            ValueError, match="contains NaN, infinity or a value too large"
        ):
            estimator.fit(X, y)
        return
    elif os.environ["PYTEST_CURRENT_TEST"].find("check_regressor_multioutput") != -1:
        # sklearn requires float64
        with pytest.raises(AssertionError, match="Got float32"):
            check(estimator)
    else:
        check(estimator)


def test_categorical(tmp_path: Path) -> None:
    X, y = tm.make_categorical(n_samples=32, n_features=2, n_categories=3, onehot=False)
    ft = ["c"] * X.shape[1]
    reg = xgb.XGBRegressor(
        feature_types=ft,
        max_cat_to_onehot=1,
    )
    reg.fit(X.values, y, eval_set=[(X.values, y)])
    from_cat = reg.evals_result()["validation_0"]["rmse"]
    predt_cat = reg.predict(X.values)
    assert reg.get_booster().feature_types == ft
    path = tmp_path / "model.json"
    reg.save_model(path)
    reg = xgb.XGBRegressor()
    reg.load_model(path)
    assert reg.feature_types == ft
    assert reg.enable_categorical is True

    onehot, y = tm.make_categorical(
        n_samples=32, n_features=2, n_categories=3, onehot=True
    )
    reg = xgb.XGBRegressor()
    reg.fit(onehot, y, eval_set=[(onehot, y)])
    from_enc = reg.evals_result()["validation_0"]["rmse"]
    predt_enc = reg.predict(onehot)

    np.testing.assert_allclose(from_cat, from_enc)
    np.testing.assert_allclose(predt_cat, predt_enc)


def test_evaluation_metric():
    from sklearn.datasets import load_diabetes, load_digits
    from sklearn.metrics import mean_absolute_error

    X, y = load_diabetes(return_X_y=True)
    n_estimators = 16

    with tm.captured_output() as (out, err):
        reg = xgb.XGBRegressor(
            tree_method="hist",
            eval_metric=mean_absolute_error,
            n_estimators=n_estimators,
        )
        reg.fit(X, y, eval_set=[(X, y)])
        lines = out.getvalue().strip().split("\n")

    assert len(lines) == n_estimators
    for line in lines:
        assert line.find("mean_absolute_error") != -1

    def merror(y_true: np.ndarray, predt: np.ndarray):
        n_samples = y_true.shape[0]
        assert n_samples == predt.size
        errors = np.zeros(y_true.shape[0])
        errors[y != predt] = 1.0
        return np.sum(errors) / n_samples

    X, y = load_digits(n_class=10, return_X_y=True)

    clf = xgb.XGBClassifier(
        tree_method="hist",
        eval_metric=merror,
        n_estimators=16,
        objective="multi:softmax",
    )
    clf.fit(X, y, eval_set=[(X, y)])
    custom = clf.evals_result()

    clf = xgb.XGBClassifier(
        tree_method="hist",
        eval_metric="merror",
        n_estimators=16,
        objective="multi:softmax",
    )
    clf.fit(X, y, eval_set=[(X, y)])
    internal = clf.evals_result()

    np.testing.assert_allclose(
        custom["validation_0"]["merror"], internal["validation_0"]["merror"], atol=1e-6
    )

    clf = xgb.XGBRFClassifier(
        tree_method="hist",
        n_estimators=16,
        objective=tm.softprob_obj(10),
        eval_metric=merror,
    )
    with pytest.raises(AssertionError):
        # shape check inside the `merror` function
        clf.fit(X, y, eval_set=[(X, y)])


def test_verbose_default_consistency() -> None:
    import inspect

    from xgboost import (
        XGBClassifier,
        XGBRanker,
        XGBRegressor,
        XGBRFClassifier,
        XGBRFRegressor,
    )

    clf_default = inspect.signature(XGBClassifier.fit).parameters["verbose"].default
    reg_default = inspect.signature(XGBRegressor.fit).parameters["verbose"].default
    rank_default = inspect.signature(XGBRanker.fit).parameters["verbose"].default
    rfc_default = inspect.signature(XGBRFClassifier.fit).parameters["verbose"].default
    rfr_default = inspect.signature(XGBRFRegressor.fit).parameters["verbose"].default

    assert clf_default == reg_default == rank_default == rfc_default == rfr_default


def test_mixed_metrics() -> None:
    from sklearn.datasets import make_classification
    from sklearn.metrics import hamming_loss, hinge_loss, log_loss

    X, y = make_classification(random_state=2025)

    clf = xgb.XGBClassifier(eval_metric=["logloss", hinge_loss], n_estimators=2)
    clf.fit(X, y, eval_set=[(X, y)])
    results = clf.evals_result()["validation_0"]
    assert "logloss" in results
    assert "hinge_loss" in results

    clf = xgb.XGBClassifier(eval_metric=[hamming_loss, log_loss], n_estimators=2)
    with pytest.raises(
        NotImplementedError, match="multiple custom metrics is not yet supported."
    ):
        clf.fit(X, y, eval_set=[(X, y)])

    clf = xgb.XGBClassifier(eval_metric=[123, log_loss], n_estimators=2)
    with pytest.raises(TypeError, match="Invalid type for the `eval_metric`"):
        clf.fit(X, y, eval_set=[(X, y)])


def test_weighted_evaluation_metric():
    from sklearn.datasets import make_hastie_10_2
    from sklearn.metrics import log_loss

    X, y = make_hastie_10_2(n_samples=2000, random_state=42)
    labels, y = np.unique(y, return_inverse=True)
    X_train, X_test = X[:1600], X[1600:]
    y_train, y_test = y[:1600], y[1600:]
    weights_eval_set = np.random.choice([1, 2], len(X_test))

    np.random.seed(0)
    weights_train = np.random.choice([1, 2], len(X_train))

    clf = xgb.XGBClassifier(
        tree_method="hist",
        eval_metric=log_loss,
        n_estimators=16,
        objective="binary:logistic",
    )
    clf.fit(
        X_train,
        y_train,
        sample_weight=weights_train,
        eval_set=[(X_test, y_test)],
        sample_weight_eval_set=[weights_eval_set],
    )
    custom = clf.evals_result()

    clf = xgb.XGBClassifier(
        tree_method="hist",
        eval_metric="logloss",
        n_estimators=16,
        objective="binary:logistic",
    )
    clf.fit(
        X_train,
        y_train,
        sample_weight=weights_train,
        eval_set=[(X_test, y_test)],
        sample_weight_eval_set=[weights_eval_set],
    )
    internal = clf.evals_result()

    np.testing.assert_allclose(
        custom["validation_0"]["log_loss"],
        internal["validation_0"]["logloss"],
        atol=1e-6,
    )


def test_intercept() -> None:
    run_intercept("cpu")


def test_fit_none() -> None:
    with pytest.raises(TypeError, match="NoneType"):
        xgb.XGBClassifier().fit(None, [0, 1])

    X = rng.normal(size=4).reshape(2, 2)

    with pytest.raises(ValueError, match="Invalid classes"):
        xgb.XGBClassifier().fit(X, None)

    with pytest.raises(ValueError, match="labels"):
        xgb.XGBRegressor().fit(X, None)


def test_tags() -> None:
    for reg in [xgb.XGBRegressor(), xgb.XGBRFRegressor()]:
        tags = reg._more_tags()
        assert "non_deterministic" not in tags
        assert tags["multioutput"] is True
        assert tags["multioutput_only"] is False

    for clf in [xgb.XGBClassifier(), xgb.XGBRFClassifier()]:
        tags = clf._more_tags()
        assert "multioutput" not in tags
        assert tags["multilabel"] is True

    tags = xgb.XGBRanker()._more_tags()
    assert "multioutput" not in tags


# the try-excepts in this test should be removed once xgboost's
# minimum supported scikit-learn version is at least 1.6
def test_sklearn_tags():
    def _assert_has_xgbmodel_tags(tags):
        # values set by XGBModel.__sklearn_tags__()
        assert tags.non_deterministic is False
        assert tags.no_validation is True
        assert tags.input_tags.allow_nan is True

    for reg in [xgb.XGBRegressor(), xgb.XGBRFRegressor()]:
        try:
            # if no AttributeError was thrown, we must be using scikit-learn>=1.6,
            # and so the actual effects of __sklearn_tags__() should be tested
            tags = reg.__sklearn_tags__()
            _assert_has_xgbmodel_tags(tags)
            # regressor-specific values
            assert tags.estimator_type == "regressor"
            assert tags.regressor_tags is not None
            assert tags.classifier_tags is None
            assert tags.target_tags.multi_output is True
            assert tags.target_tags.single_output is True
        except AttributeError as err:
            # only the exact error we expected to be raised should be raised
            assert bool(re.search(r"__sklearn_tags__.* should not be called", str(err)))

    for clf in [xgb.XGBClassifier(), xgb.XGBRFClassifier()]:
        try:
            # if no AttributeError was thrown, we must be using scikit-learn>=1.6,
            # and so the actual effects of __sklearn_tags__() should be tested
            tags = clf.__sklearn_tags__()
            _assert_has_xgbmodel_tags(tags)
            # classifier-specific values
            assert tags.estimator_type == "classifier"
            assert tags.regressor_tags is None
            assert tags.classifier_tags is not None
            assert tags.classifier_tags.multi_label is True
        except AttributeError as err:
            # only the exact error we expected to be raised should be raised
            assert bool(re.search(r"__sklearn_tags__.* should not be called", str(err)))

    for rnk in [xgb.XGBRanker()]:
        try:
            # if no AttributeError was thrown, we must be using scikit-learn>=1.6,
            # and so the actual effects of __sklearn_tags__() should be tested
            tags = rnk.__sklearn_tags__()
            _assert_has_xgbmodel_tags(tags)
        except AttributeError as err:
            # only the exact error we expected to be raised should be raised
            assert bool(re.search(r"__sklearn_tags__.* should not be called", str(err)))


def test_doc_link() -> None:
    for est in [
        xgb.XGBRegressor(),
        xgb.XGBClassifier(),
        xgb.XGBRanker(),
        xgb.XGBRFRegressor(),
        xgb.XGBRFClassifier(),
    ]:
        name = est.__class__.__name__
        link = est._get_doc_link()
        assert f"xgboost.{name}" in link


def test_apply_method() -> None:
    import pandas as pd

    X_num = np.random.rand(5, 5)
    df = pd.DataFrame(X_num, columns=[f"f{i}" for i in range(X_num.shape[1])])
    df["test"] = pd.Series(
        ["one", "two", "three", "four", "five"], dtype="category"
    )  # <- categorical column
    y = np.arange(len(df))

    model = xgb.XGBClassifier()
    model.fit(df, y)

    model.apply(df)  # this must not raise

    model.set_params(enable_categorical=False)
    with pytest.raises(ValueError, match="`enable_categorical`"):
        model.apply(df)


def test_recoding() -> None:
    run_recoding("cpu")



import os
import subprocess

import pytest

from xgboost import testing as tm

pytestmark = [
    pytest.mark.skipif(**tm.no_dask()),
    pytest.mark.skipif(**tm.no_dask_cuda()),
    tm.timeout(60),
]


@pytest.mark.skipif(**tm.no_cupy())
@pytest.mark.mgpu
def test_dask_training() -> None:
    script = os.path.join(tm.demo_dir(__file__), "dask", "gpu_training.py")
    cmd = ["python", script]
    subprocess.check_call(cmd)


@pytest.mark.mgpu
def test_dask_sklearn_demo() -> None:
    script = os.path.join(tm.demo_dir(__file__), "dask", "sklearn_gpu_training.py")
    cmd = ["python", script]
    subprocess.check_call(cmd)


@pytest.mark.mgpu
@pytest.mark.skipif(**tm.no_cupy())
def test_forward_logging_demo() -> None:
    script = os.path.join(tm.demo_dir(__file__), "dask", "forward_logging.py")
    cmd = ["python", script]
    subprocess.check_call(cmd)



import json
import os
import pickle

import numpy as np

import xgboost as xgb

kRows = 100
kCols = 10


def generate_data():
    X = np.random.randn(kRows, kCols)
    y = np.random.randn(kRows)
    return X, y


class TestPickling:
    def run_model_pickling(self, xgb_params) -> str:
        X, y = generate_data()
        dtrain = xgb.DMatrix(X, y)
        bst = xgb.train(xgb_params, dtrain)

        dump_0 = bst.get_dump(dump_format="json")
        assert dump_0
        config_0 = bst.save_config()

        filename = "model.pkl"

        with open(filename, "wb") as fd:
            pickle.dump(bst, fd)

        with open(filename, "rb") as fd:
            bst = pickle.load(fd)

        with open(filename, "wb") as fd:
            pickle.dump(bst, fd)

        with open(filename, "rb") as fd:
            bst = pickle.load(fd)

        assert bst.get_dump(dump_format="json") == dump_0

        if os.path.exists(filename):
            os.remove(filename)

        config_1 = bst.save_config()
        assert config_0 == config_1
        return json.loads(config_0)

    def test_model_pickling_json(self):
        def check(config):
            tree_param = config["learner"]["gradient_booster"]["tree_train_param"]
            subsample = tree_param["subsample"]
            assert float(subsample) == 0.5

        params = {"nthread": 8, "tree_method": "hist", "subsample": 0.5}
        config = self.run_model_pickling(params)
        check(config)
        params = {"nthread": 8, "tree_method": "exact", "subsample": 0.5}
        config = self.run_model_pickling(params)
        check(config)



import os
import subprocess
from pathlib import Path

import numpy as np
import pytest
import xgboost as xgb
from xgboost import testing as tm

pytestmark = tm.timeout(10)


class TestOMP:
    def test_omp(self):
        dtrain, dtest = tm.load_agaricus(__file__)

        param = {
            "booster": "gbtree",
            "objective": "binary:logistic",
            "grow_policy": "depthwise",
            "tree_method": "hist",
            "eval_metric": "error",
            "max_depth": 5,
            "min_child_weight": 0,
        }

        watchlist = [(dtest, "eval"), (dtrain, "train")]
        num_round = 5

        def run_trial():
            res = {}
            bst = xgb.train(param, dtrain, num_round, watchlist, evals_result=res)
            metrics = [res["train"]["error"][-1], res["eval"]["error"][-1]]
            preds = bst.predict(dtest)
            return metrics, preds

        def consist_test(title, n):
            auc, pred = run_trial()
            for i in range(n - 1):
                auc2, pred2 = run_trial()
                try:
                    assert auc == auc2
                    assert np.array_equal(pred, pred2)
                except Exception as e:
                    print("-------test %s failed, num_trial: %d-------" % (title, i))
                    raise e
                auc, pred = auc2, pred2
            return auc, pred

        print("test approx ...")
        param["tree_method"] = "approx"

        n_trials = 10
        param["nthread"] = 1
        auc_1, pred_1 = consist_test("approx_thread_1", n_trials)

        param["nthread"] = 2
        auc_2, pred_2 = consist_test("approx_thread_2", n_trials)

        param["nthread"] = 3
        auc_3, pred_3 = consist_test("approx_thread_3", n_trials)

        assert auc_1 == auc_2 == auc_3
        assert np.array_equal(auc_1, auc_2)
        assert np.array_equal(auc_1, auc_3)

        print("test hist ...")
        param["tree_method"] = "hist"

        param["nthread"] = 1
        auc_1, pred_1 = consist_test("hist_thread_1", n_trials)

        param["nthread"] = 2
        auc_2, pred_2 = consist_test("hist_thread_2", n_trials)

        param["nthread"] = 3
        auc_3, pred_3 = consist_test("hist_thread_3", n_trials)

        assert auc_1 == auc_2 == auc_3
        assert np.array_equal(auc_1, auc_2)
        assert np.array_equal(auc_1, auc_3)

    @pytest.mark.skipif(**tm.no_sklearn())
    @pytest.mark.timeout(30)
    def test_with_omp_thread_limit(self, tmp_path: Path) -> None:
        args = [
            "python",
            os.path.join(os.path.dirname(tm.normpath(__file__)), "with_omp_limit.py"),
        ]
        results = []
        for i in (1, 2, 16):
            path = tmp_path / str(i)
            with open(path, "w") as fd:
                fd.write("\n")
            cp = args.copy()
            cp.append(str(path))

            env = os.environ.copy()
            env["OMP_THREAD_LIMIT"] = str(i)

            status = subprocess.call(cp, env=env)
            assert status == 0

            with open(path, "r") as fd:
                results.append(float(fd.read()))

        for auc in results:
            np.testing.assert_allclose(auc, results[0])



from collections import namedtuple
from pathlib import Path
from typing import Optional, Tuple, Union

import numpy as np
import pytest
import xgboost as xgb
from xgboost import testing as tm
from xgboost.testing.callbacks import (
    run_eta_decay,
    run_eta_decay_leaf_output,
    tree_methods_objs,
)

# We use the dataset for tests.
pytestmark = pytest.mark.skipif(**tm.no_sklearn())


BreastCancer = namedtuple("BreastCancer", ["full", "tr", "va"])


@pytest.fixture
def breast_cancer() -> BreastCancer:
    from sklearn.datasets import load_breast_cancer

    X, y = load_breast_cancer(return_X_y=True)

    split = int(X.shape[0] * 0.8)
    return BreastCancer(
        full=(X, y),
        tr=(X[:split, ...], y[:split, ...]),
        va=(X[split:, ...], y[split:, ...]),
    )


def eval_error_metric(predt: np.ndarray, dtrain: xgb.DMatrix) -> Tuple[str, np.float64]:
    # No custom objective, receive transformed output
    return tm.eval_error_metric(predt, dtrain, rev_link=False)


class TestCallbacks:
    def run_evaluation_monitor(
        self,
        D_train: xgb.DMatrix,
        D_valid: xgb.DMatrix,
        rounds: int,
        verbose_eval: Union[bool, int],
    ):
        def check_output(output: str) -> None:
            if int(verbose_eval) == 1:
                # Should print each iteration info
                assert len(output.split("\n")) == rounds
            elif int(verbose_eval) > rounds:
                # Should print first and latest iteration info
                assert len(output.split("\n")) == 2
            else:
                # Should print info by each period additionaly to first and latest
                # iteration
                num_periods = rounds // int(verbose_eval)
                # Extra information is required for latest iteration
                is_extra_info_required = num_periods * int(verbose_eval) < (rounds - 1)
                assert len(output.split("\n")) == (
                    1 + num_periods + int(is_extra_info_required)
                )

        evals_result: xgb.callback.TrainingCallback.EvalsLog = {}
        params = {"objective": "binary:logistic", "eval_metric": "error"}
        with tm.captured_output() as (out, err):
            xgb.train(
                params,
                D_train,
                evals=[(D_train, "Train"), (D_valid, "Valid")],
                num_boost_round=rounds,
                evals_result=evals_result,
                verbose_eval=verbose_eval,
            )
            output: str = out.getvalue().strip()
            check_output(output)

        with tm.captured_output() as (out, err):
            xgb.cv(params, D_train, num_boost_round=rounds, verbose_eval=verbose_eval)
            output = out.getvalue().strip()
            check_output(output)

    def test_evaluation_monitor(self, breast_cancer: BreastCancer) -> None:
        D_train = xgb.DMatrix(breast_cancer.tr[0], breast_cancer.tr[1])
        D_valid = xgb.DMatrix(breast_cancer.va[0], breast_cancer.va[1])
        evals_result = {}
        rounds = 10
        xgb.train(
            {"objective": "binary:logistic", "eval_metric": "error"},
            D_train,
            evals=[(D_train, "Train"), (D_valid, "Valid")],
            num_boost_round=rounds,
            evals_result=evals_result,
            verbose_eval=True,
        )
        assert len(evals_result["Train"]["error"]) == rounds
        assert len(evals_result["Valid"]["error"]) == rounds

        self.run_evaluation_monitor(D_train, D_valid, rounds, True)
        self.run_evaluation_monitor(D_train, D_valid, rounds, 2)
        self.run_evaluation_monitor(D_train, D_valid, rounds, 4)
        self.run_evaluation_monitor(D_train, D_valid, rounds, rounds + 1)

    def test_early_stopping(self, breast_cancer: BreastCancer) -> None:
        D_train = xgb.DMatrix(breast_cancer.tr[0], breast_cancer.tr[1])
        D_valid = xgb.DMatrix(breast_cancer.va[0], breast_cancer.va[1])
        evals_result = {}
        rounds = 30
        early_stopping_rounds = 5
        booster = xgb.train(
            {"objective": "binary:logistic", "eval_metric": "error"},
            D_train,
            evals=[(D_train, "Train"), (D_valid, "Valid")],
            num_boost_round=rounds,
            evals_result=evals_result,
            verbose_eval=True,
            early_stopping_rounds=early_stopping_rounds,
        )
        dump = booster.get_dump(dump_format="json")
        assert len(dump) - booster.best_iteration == early_stopping_rounds + 1

    def test_early_stopping_custom_eval(self, breast_cancer: BreastCancer) -> None:
        D_train = xgb.DMatrix(breast_cancer.tr[0], breast_cancer.tr[1])
        D_valid = xgb.DMatrix(breast_cancer.va[0], breast_cancer.va[1])
        early_stopping_rounds = 5
        booster = xgb.train(
            {
                "objective": "binary:logistic",
                "eval_metric": "error",
                "tree_method": "hist",
            },
            D_train,
            evals=[(D_train, "Train"), (D_valid, "Valid")],
            custom_metric=eval_error_metric,
            num_boost_round=1000,
            early_stopping_rounds=early_stopping_rounds,
            verbose_eval=False,
        )
        dump = booster.get_dump(dump_format="json")
        assert len(dump) - booster.best_iteration == early_stopping_rounds + 1

    def test_early_stopping_customize(self, breast_cancer: BreastCancer) -> None:
        D_train = xgb.DMatrix(breast_cancer.tr[0], breast_cancer.tr[1])
        D_valid = xgb.DMatrix(breast_cancer.va[0], breast_cancer.va[1])
        early_stopping_rounds = 5
        early_stop = xgb.callback.EarlyStopping(
            rounds=early_stopping_rounds, metric_name="CustomErr", data_name="Train"
        )
        # Specify which dataset and which metric should be used for early stopping.
        booster = xgb.train(
            {
                "objective": "binary:logistic",
                "eval_metric": ["error", "rmse"],
                "tree_method": "hist",
            },
            D_train,
            evals=[(D_train, "Train"), (D_valid, "Valid")],
            custom_metric=eval_error_metric,
            num_boost_round=1000,
            callbacks=[early_stop],
            verbose_eval=False,
        )
        dump = booster.get_dump(dump_format="json")
        assert len(dump) - booster.best_iteration == early_stopping_rounds + 1
        assert len(early_stop.stopping_history["Train"]["CustomErr"]) == len(dump)

        rounds = 100
        early_stop = xgb.callback.EarlyStopping(
            rounds=early_stopping_rounds,
            metric_name="CustomErr",
            data_name="Train",
            min_delta=100,
            save_best=True,
        )
        booster = xgb.train(
            {
                "objective": "binary:logistic",
                "eval_metric": ["error", "rmse"],
                "tree_method": "hist",
            },
            D_train,
            evals=[(D_train, "Train"), (D_valid, "Valid")],
            # No custom objective, transformed output
            custom_metric=eval_error_metric,
            num_boost_round=rounds,
            callbacks=[early_stop],
            verbose_eval=False,
        )
        # No iteration can be made with min_delta == 100
        assert booster.best_iteration == 0
        assert booster.num_boosted_rounds() == 1

    def test_early_stopping_skl(self, breast_cancer: BreastCancer) -> None:
        X, y = breast_cancer.full
        early_stopping_rounds = 5
        cls = xgb.XGBClassifier(
            early_stopping_rounds=early_stopping_rounds, eval_metric="error"
        )
        cls.fit(X, y, eval_set=[(X, y)])
        booster = cls.get_booster()
        dump = booster.get_dump(dump_format="json")
        assert len(dump) - booster.best_iteration == early_stopping_rounds + 1

    def test_early_stopping_custom_eval_skl(self, breast_cancer: BreastCancer) -> None:
        X, y = breast_cancer.full
        early_stopping_rounds = 5
        early_stop = xgb.callback.EarlyStopping(rounds=early_stopping_rounds)
        cls = xgb.XGBClassifier(
            eval_metric=tm.eval_error_metric_skl, callbacks=[early_stop]
        )
        cls.fit(X, y, eval_set=[(X, y)])
        booster = cls.get_booster()
        dump = booster.get_dump(dump_format="json")
        assert len(dump) - booster.best_iteration == early_stopping_rounds + 1

    def test_early_stopping_save_best_model(self, breast_cancer: BreastCancer) -> None:
        X, y = breast_cancer.full
        n_estimators = 100
        early_stopping_rounds = 5
        early_stop = xgb.callback.EarlyStopping(
            rounds=early_stopping_rounds, save_best=True
        )
        cls = xgb.XGBClassifier(
            n_estimators=n_estimators,
            eval_metric=tm.eval_error_metric_skl,
            callbacks=[early_stop],
        )
        cls.fit(X, y, eval_set=[(X, y)])
        booster = cls.get_booster()
        dump = booster.get_dump(dump_format="json")
        assert len(dump) == booster.best_iteration + 1

        early_stop = xgb.callback.EarlyStopping(
            rounds=early_stopping_rounds, save_best=True
        )
        cls = xgb.XGBClassifier(
            booster="gblinear",
            n_estimators=10,
            eval_metric=tm.eval_error_metric_skl,
            callbacks=[early_stop],
        )
        with pytest.raises(ValueError):
            cls.fit(X, y, eval_set=[(X, y)])

        # No error
        early_stop = xgb.callback.EarlyStopping(
            rounds=early_stopping_rounds, save_best=False
        )
        xgb.XGBClassifier(
            booster="gblinear",
            n_estimators=10,
            eval_metric=tm.eval_error_metric_skl,
            callbacks=[early_stop],
        ).fit(X, y, eval_set=[(X, y)])

    def test_early_stopping_continuation(
        self, breast_cancer: BreastCancer, tmp_path: Path
    ) -> None:
        X, y = breast_cancer.full

        early_stopping_rounds = 5
        early_stop = xgb.callback.EarlyStopping(
            rounds=early_stopping_rounds, save_best=True
        )
        cls = xgb.XGBClassifier(
            eval_metric=tm.eval_error_metric_skl, callbacks=[early_stop]
        )
        cls.fit(X, y, eval_set=[(X, y)])

        booster = cls.get_booster()
        assert booster.num_boosted_rounds() == booster.best_iteration + 1

        path = tmp_path / "model.json"
        cls.save_model(path)
        cls = xgb.XGBClassifier()
        cls.load_model(path)
        assert cls._Booster is not None
        early_stopping_rounds = 3
        cls.set_params(
            eval_metric=tm.eval_error_metric_skl,
            early_stopping_rounds=early_stopping_rounds,
        )
        cls.fit(X, y, eval_set=[(X, y)])
        booster = cls.get_booster()
        assert (
            booster.num_boosted_rounds()
            == booster.best_iteration + early_stopping_rounds + 1
        )

    def test_early_stopping_multiple_metrics(self):
        from sklearn.datasets import make_classification

        X, y = make_classification(random_state=1994)
        # AUC approaches 1.0 real quick.
        clf = xgb.XGBClassifier(eval_metric=["logloss", "auc"], early_stopping_rounds=2)
        clf.fit(X, y, eval_set=[(X, y)])
        assert clf.best_iteration < 8
        assert clf.evals_result()["validation_0"]["auc"][-1] > 0.99

        clf = xgb.XGBClassifier(eval_metric=["auc", "logloss"], early_stopping_rounds=2)
        clf.fit(X, y, eval_set=[(X, y)])

        assert clf.best_iteration > 50
        assert clf.evals_result()["validation_0"]["auc"][-1] > 0.99

    @pytest.mark.parametrize("tree_method", ["hist", "approx", "exact"])
    def test_eta_decay(self, tree_method: str) -> None:
        dtrain, dtest = tm.load_agaricus(__file__)
        run_eta_decay(tree_method, dtrain, dtest, "cpu")

    @pytest.mark.parametrize("tree_method,objective", tree_methods_objs())
    def test_eta_decay_leaf_output(self, tree_method: str, objective: str) -> None:
        dtrain, dtest = tm.load_agaricus(__file__)
        run_eta_decay_leaf_output(tree_method, objective, dtrain, dtest, "cpu")

    def test_check_point(self, breast_cancer: BreastCancer, tmp_path: Path) -> None:
        X, y = breast_cancer.full
        m = xgb.DMatrix(X, y)
        check_point = xgb.callback.TrainingCheckPoint(
            directory=tmp_path, interval=1, name="model"
        )
        xgb.train(
            {"objective": "binary:logistic"},
            m,
            num_boost_round=10,
            verbose_eval=False,
            callbacks=[check_point],
        )
        for i in range(1, 10):
            assert (
                tmp_path / f"model_{i}.{xgb.callback.TrainingCheckPoint.default_format}"
            ).exists()

        check_point = xgb.callback.TrainingCheckPoint(
            directory=tmp_path, interval=1, as_pickle=True, name="model"
        )
        xgb.train(
            {"objective": "binary:logistic"},
            m,
            num_boost_round=10,
            verbose_eval=False,
            callbacks=[check_point],
        )
        for i in range(1, 10):
            assert (tmp_path / f"model_{i}.pkl").exists()

    def test_callback_list(self) -> None:
        X, y = tm.data.get_california_housing()
        m = xgb.DMatrix(X, y)
        callbacks = [xgb.callback.EarlyStopping(rounds=10)]
        for i in range(4):
            xgb.train(
                {"objective": "reg:squarederror", "eval_metric": "rmse"},
                m,
                evals=[(m, "Train")],
                num_boost_round=1,
                verbose_eval=True,
                callbacks=callbacks,
            )
        assert len(callbacks) == 1

    def test_attribute_error(self, breast_cancer: BreastCancer) -> None:
        X, y = breast_cancer.full

        clf = xgb.XGBClassifier(n_estimators=8)
        clf.fit(X, y, eval_set=[(X, y)])

        with pytest.raises(AttributeError, match="early stopping is used"):
            clf.best_iteration

        with pytest.raises(AttributeError, match="early stopping is used"):
            clf.best_score

        booster = clf.get_booster()
        with pytest.raises(AttributeError, match="early stopping is used"):
            booster.best_iteration

        with pytest.raises(AttributeError, match="early stopping is used"):
            booster.best_score

    def test_preserve_order(self) -> None:
        """Test the ordering of the callbacks is preserved."""
        X, y, w = tm.make_regression(256, 16, False)
        fst_call: Optional[int] = None

        # If we use Python `set`, Cb1 is ordered before Cb2. This test makes sure Cb2 is
        # called before Cb1.
        class Cb2(xgb.callback.TrainingCallback):
            def before_iteration(self, model, epoch: int, evals_log) -> bool:
                nonlocal fst_call
                assert fst_call is None or fst_call == 2
                fst_call = 2
                return False

        class Cb1(xgb.callback.TrainingCallback):
            def before_iteration(self, model, epoch: int, evals_log) -> bool:
                assert fst_call == 2
                return False

        callbacks = [Cb2(), Cb1()]
        xgb.train({}, dtrain=xgb.QuantileDMatrix(X, y, weight=w), callbacks=callbacks)



import socket
from dataclasses import asdict

import numpy as np
import pytest

import xgboost as xgb
from xgboost import RabitTracker, build_info, federated
from xgboost import testing as tm
from xgboost.collective import Config


def run_rabit_worker(rabit_env: dict, world_size: int) -> int:
    with xgb.collective.CommunicatorContext(**rabit_env):
        assert xgb.collective.get_world_size() == world_size
        assert xgb.collective.is_distributed()
        assert xgb.collective.get_processor_name() == socket.gethostname()
        ret = xgb.collective.broadcast("test1234", 0)
        assert str(ret) == "test1234"
        reduced = xgb.collective.allreduce(np.asarray([1, 2, 3]), xgb.collective.Op.SUM)
        assert np.array_equal(reduced, np.asarray([2, 4, 6]))
    return 0


@pytest.mark.skipif(**tm.no_loky())
def test_rabit_communicator() -> None:
    from loky import get_reusable_executor

    world_size = 2
    tracker = RabitTracker(host_ip="127.0.0.1", n_workers=world_size)
    tracker.start()
    workers = []

    with get_reusable_executor(max_workers=world_size) as pool:
        for _ in range(world_size):
            worker = pool.submit(
                run_rabit_worker, rabit_env=tracker.worker_args(), world_size=world_size
            )
            workers.append(worker)

        for worker in workers:
            assert worker.result() == 0


def run_federated_worker(port: int, world_size: int, rank: int) -> int:
    with xgb.collective.CommunicatorContext(
        dmlc_communicator="federated",
        federated_server_address=f"localhost:{port}",
        federated_world_size=world_size,
        federated_rank=rank,
    ):
        assert xgb.collective.get_world_size() == world_size
        assert xgb.collective.is_distributed()
        assert xgb.collective.get_processor_name() == f"rank:{rank}"
        bret = xgb.collective.broadcast("test1234", 0)
        assert str(bret) == "test1234"
        aret = xgb.collective.allreduce(np.asarray([1, 2, 3]), xgb.collective.Op.SUM)
        assert np.array_equal(aret, np.asarray([2, 4, 6]))
    return 0


@pytest.mark.skipif(**tm.skip_win())
@pytest.mark.skipif(**tm.no_loky())
def test_federated_communicator() -> None:
    from loky import get_reusable_executor

    if not build_info()["USE_FEDERATED"]:
        pytest.skip("XGBoost not built with federated learning enabled")

    port = 9091
    world_size = 2
    with get_reusable_executor(max_workers=world_size + 1) as pool:
        kwargs = {"port": port, "n_workers": world_size, "blocking": False}
        tracker = pool.submit(federated.run_federated_server, **kwargs)
        if not tracker.running():
            raise RuntimeError("Error starting Federated Learning server")

        workers = []
        for rank in range(world_size):
            worker = pool.submit(
                run_federated_worker, port=port, world_size=world_size, rank=rank
            )
            workers.append(worker)
        for worker in workers:
            assert worker.result() == 0


def test_config_serialization() -> None:
    cfg = Config(retry=1, timeout=2, tracker_host_ip="127.0.0.1", tracker_port=None)
    cfg1 = Config(**asdict(cfg))
    assert cfg == cfg1



import os
from concurrent.futures import ThreadPoolExecutor
from typing import Type

import numpy as np
import pytest
import xgboost as xgb
from xgboost import testing as tm
from xgboost.testing.data import make_categorical
from xgboost.testing.ordinal import (
    run_basic_predict,
    run_cat_container,
    run_cat_container_iter,
    run_cat_container_mixed,
    run_cat_invalid,
    run_cat_leaf,
    run_cat_oov_in_range,
    run_cat_predict,
    run_cat_shap,
    run_cat_thread_safety,
    run_recode_dmatrix,
    run_recode_dmatrix_predict,
    run_specified_cat,
    run_training_continuation,
    run_update,
    run_validation,
)

pytestmark = pytest.mark.skipif(**tm.no_multiple(tm.no_arrow(), tm.no_cudf()))


def test_cat_container() -> None:
    run_cat_container("cuda")


def test_cat_container_mixed() -> None:
    run_cat_container_mixed("cuda")


def test_cat_container_iter() -> None:
    run_cat_container_iter("cuda")


def test_cat_predict() -> None:
    run_cat_predict("cuda")


def test_cat_invalid() -> None:
    run_cat_invalid("cuda")


def test_cat_oov_in_range() -> None:
    run_cat_oov_in_range("cuda")


def test_cat_thread_safety() -> None:
    run_cat_thread_safety("cuda")


def test_cat_shap() -> None:
    run_cat_shap("cuda")


def test_cat_leaf() -> None:
    run_cat_leaf("cuda")


def test_mixed_devices() -> None:
    n_samples = 128
    n_features = 4
    X, y = make_categorical(n_samples, n_features, 7, onehot=False, device="cpu")

    def run_cpu_gpu(DMatrixT: Type) -> bool:
        Xy = DMatrixT(X, y)
        booster = xgb.train({"tree_method": "hist", "device": "cuda"}, Xy)
        predt0 = booster.inplace_predict(X)
        predt1 = booster.predict(DMatrixT(X, y))

        np.testing.assert_allclose(predt0, predt1)
        return True

    n_cpus = os.cpu_count()
    assert n_cpus is not None

    futures = []
    with ThreadPoolExecutor(max_workers=n_cpus) as e:
        for dm in (xgb.DMatrix, xgb.QuantileDMatrix):
            f = e.submit(run_cpu_gpu, dm)
            futures.append(f)

    for f in futures:
        assert f.result()

    X, y = make_categorical(n_samples, n_features, 7, onehot=False, device="cuda")

    def run_gpu_cpu(DMatrixT: Type) -> bool:
        Xy = DMatrixT(X, y)
        booster = xgb.train({"tree_method": "hist", "device": "cpu"}, Xy)
        p = booster.inplace_predict(X)
        assert not isinstance(p, np.ndarray)
        predt0 = p.get()
        predt1 = booster.predict(DMatrixT(X, y))

        np.testing.assert_allclose(predt0, predt1)
        return True

    futures = []
    with ThreadPoolExecutor(max_workers=n_cpus) as e:
        for dm in (xgb.DMatrix, xgb.QuantileDMatrix):
            f = e.submit(run_gpu_cpu, dm)
            futures.append(f)

    for f in futures:
        assert f.result()


@pytest.mark.parametrize("DMatrixT", [xgb.DMatrix, xgb.QuantileDMatrix])
def test_mixed_devices_types(DMatrixT: Type) -> None:
    run_basic_predict(DMatrixT, "cuda", "cpu")
    run_basic_predict(DMatrixT, "cpu", "cuda")


def test_specified_cat() -> None:
    run_specified_cat("cuda")


def test_validation() -> None:
    run_validation("cuda")


def test_recode_dmatrix() -> None:
    run_recode_dmatrix("cuda")


def test_training_continuation() -> None:
    run_training_continuation("cuda")


def test_update() -> None:
    run_update("cuda")


def test_recode_dmatrix_predict() -> None:
    run_recode_dmatrix_predict("cuda")



"""
Demo for GLM
============
"""
import os

import xgboost as xgb

##
#  this script demonstrate how to fit generalized linear model in xgboost
#  basically, we are using linear model, instead of tree for our boosters
##
CURRENT_DIR = os.path.dirname(__file__)
dtrain = xgb.DMatrix(
    os.path.join(CURRENT_DIR, "../data/agaricus.txt.train?format=libsvm")
)
dtest = xgb.DMatrix(
    os.path.join(CURRENT_DIR, "../data/agaricus.txt.test?format=libsvm")
)
# change booster to gblinear, so that we are fitting a linear model
# alpha is the L1 regularizer
# lambda is the L2 regularizer
# you can also set lambda_bias which is L2 regularizer on the bias term
param = {
    "objective": "binary:logistic",
    "booster": "gblinear",
    "alpha": 0.0001,
    "lambda": 1,
}

# normally, you do not need to set eta (step_size)
# XGBoost uses a parallel coordinate descent algorithm (shotgun),
# there could be affection on convergence with parallelization on certain cases
# setting eta to be smaller value, e.g 0.5 can make the optimization more stable
# param['eta'] = 1

##
# the rest of settings are the same
##
watchlist = [(dtest, "eval"), (dtrain, "train")]
num_round = 4
bst = xgb.train(param, dtrain, num_round, watchlist)
preds = bst.predict(dtest)
labels = dtest.get_label()
print(
    "error=%f"
    % (
        sum(1 for i in range(len(preds)) if int(preds[i] > 0.5) != labels[i])
        / float(len(preds))
    )
)



"""
A demo for multi-output regression
==================================

The demo is adopted from scikit-learn:

https://scikit-learn.org/stable/auto_examples/ensemble/plot_random_forest_regression_multioutput.html#sphx-glr-auto-examples-ensemble-plot-random-forest-regression-multioutput-py

See :doc:`/tutorials/multioutput` for more information.

.. note::

    The feature is experimental. For the `multi_output_tree` strategy, many features are
    missing.

.. seealso:: :ref:`sphx_glr_python_examples_multioutput_reduced_gradient.py`

"""

import argparse
from typing import Dict, List, Optional, Tuple

import matplotlib
import numpy as np
from matplotlib import pyplot as plt

import xgboost as xgb


def plot_predt(
    y: np.ndarray, y_predt: np.ndarray, name: str, ax: matplotlib.axes.Axes
) -> None:
    s = 25
    ax.scatter(y[:, 0], y[:, 1], c="navy", s=s, edgecolor="black", label=name)
    ax.scatter(y_predt[:, 0], y_predt[:, 1], c="cornflowerblue", s=s, edgecolor="black")
    ax.legend()


def gen_circle() -> Tuple[np.ndarray, np.ndarray]:
    "Generate a sample dataset that y is a 2 dim circle."
    rng = np.random.RandomState(1994)
    X = np.sort(200 * rng.rand(100, 1) - 100, axis=0)
    y = np.array([np.pi * np.sin(X).ravel(), np.pi * np.cos(X).ravel()]).T
    y[::5, :] += 0.5 - rng.rand(20, 2)
    y = y - y.min()
    y = y / y.max()
    return X, y


def rmse_model(strategy: str, ax: Optional[matplotlib.axes.Axes]) -> None:
    """Draw a circle with 2-dim coordinate as target variables."""
    X, y = gen_circle()
    # Train a regressor on it
    reg = xgb.XGBRegressor(
        tree_method="hist",
        n_estimators=128,
        n_jobs=16,
        max_depth=8,
        multi_strategy=strategy,
        subsample=0.6,
    )
    reg.fit(X, y, eval_set=[(X, y)])

    y_predt = reg.predict(X)
    if ax:
        plot_predt(y, y_predt, f"RMSE-{strategy}", ax)


def custom_rmse_model(strategy: str, ax: Optional[matplotlib.axes.Axes]) -> None:
    """Train using Python implementation of Squared Error."""

    def gradient(predt: np.ndarray, dtrain: xgb.DMatrix) -> np.ndarray:
        """Compute the gradient squared error."""
        y = dtrain.get_label().reshape(predt.shape)
        return predt - y

    def hessian(predt: np.ndarray, dtrain: xgb.DMatrix) -> np.ndarray:
        """Compute the hessian for squared error."""
        return np.ones(predt.shape)

    def squared_log(
        predt: np.ndarray, dtrain: xgb.DMatrix
    ) -> Tuple[np.ndarray, np.ndarray]:
        grad = gradient(predt, dtrain)
        hess = hessian(predt, dtrain)
        # both numpy.ndarray and cupy.ndarray works.
        return grad, hess

    def rmse(predt: np.ndarray, dtrain: xgb.DMatrix) -> Tuple[str, float]:
        y = dtrain.get_label().reshape(predt.shape)
        v = np.sqrt(np.mean(np.power(y - predt, 2)))
        return "PyRMSE", v

    X, y = gen_circle()
    Xy = xgb.DMatrix(X, y)
    results: Dict[str, Dict[str, List[float]]] = {}
    # Make sure the `num_target` is passed to XGBoost when custom objective is used.
    # When builtin objective is used, XGBoost can figure out the number of targets
    # automatically.
    booster = xgb.train(
        {
            "tree_method": "hist",
            "num_target": y.shape[1],
            "multi_strategy": strategy,
        },
        dtrain=Xy,
        num_boost_round=128,
        obj=squared_log,
        evals=[(Xy, "Train")],
        evals_result=results,
        custom_metric=rmse,
    )

    y_predt = booster.inplace_predict(X)
    if ax:
        plot_predt(y, y_predt, f"PyRMSE-{strategy}", ax)

    np.testing.assert_allclose(
        results["Train"]["rmse"], results["Train"]["PyRMSE"], rtol=1e-2
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--plot", choices=[0, 1], type=int, default=1)
    args = parser.parse_args()
    if args.plot == 1:
        _, axs = plt.subplots(2, 2)
    else:
        axs = np.full(shape=(2, 2), fill_value=None)
    assert isinstance(axs, np.ndarray)

    # Train with builtin RMSE objective
    # - One model per output.
    rmse_model("one_output_per_tree", axs[0, 0])
    # - One model for all outputs, this is still working in progress, many features are
    # missing.
    rmse_model("multi_output_tree", axs[0, 1])

    # Train with custom objective.
    # - One model per output.
    custom_rmse_model("one_output_per_tree", axs[1, 0])
    # - One model for all outputs, this is still working in progress, many features are
    # missing.
    custom_rmse_model("multi_output_tree", axs[1, 1])
    if args.plot == 1:
        plt.show()



"""Cross-platform model test: Train on GPU (Linux), test inference on macOS."""

import argparse
import pickle
import sys
from pathlib import Path
from typing import Tuple

import numpy as np
import xgboost as xgb
from sklearn.datasets import make_classification
from sklearn.metrics import accuracy_score

SEED = 2026


def _pickle_path(model_path: str) -> Path:
    return Path(model_path).with_suffix(".pkl")


def get_data() -> Tuple[np.ndarray, np.ndarray]:
    """Generate reproducible synthetic classification data."""
    X, y = make_classification(
        n_samples=1000,
        n_features=20,
        n_informative=10,
        n_classes=3,
        n_clusters_per_class=1,
        random_state=SEED,
    )
    return X.astype(np.float32), y.astype(np.int32)


def train_model(model_path: str) -> None:
    """Train models using GPU and save them (binary + pickle with column sampling)."""
    X, y = get_data()

    clf = xgb.XGBClassifier(
        device="cuda",
        n_estimators=50,
        max_depth=6,
        learning_rate=0.3,
        random_state=SEED,
        colsample_bynode=0.8,
    )
    clf.fit(X, y)

    accuracy = accuracy_score(y, clf.predict(X))
    clf.get_booster().set_attr(expected_accuracy=str(accuracy))
    clf.save_model(model_path)

    with open(_pickle_path(model_path), "wb") as fd:
        pickle.dump(clf.get_booster(), fd)


def test_inference(model_path: str) -> None:
    """Load models and verify predictions match (binary + pickle)."""
    X, y = get_data()

    clf = xgb.XGBClassifier()
    clf.load_model(model_path)

    accuracy = accuracy_score(y, clf.predict(X))
    ea = clf.get_booster().attr("expected_accuracy")
    assert ea is not None
    expected_accuracy = float(ea)
    np.testing.assert_allclose(accuracy, expected_accuracy)

    with open(_pickle_path(model_path), "rb") as f:
        booster = pickle.load(f)

    clf = xgb.XGBClassifier(n_estimators=2)
    clf.fit(X, y, xgb_model=booster)


def main() -> int:
    """Entry for both training and inference."""
    parser = argparse.ArgumentParser(description="Cross-platform XGBoost model test.")
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("--train", action="store_true", help="Train models using GPU")
    group.add_argument("--inference", action="store_true", help="Test inference")
    parser.add_argument(
        "--model-path",
        type=str,
        default="cross_platform_model.ubj",
        help="Path to model file (pickle path is derived by replacing extension)",
    )

    args = parser.parse_args()

    if args.train:
        train_model(args.model_path)
    else:
        test_inference(args.model_path)

    return 0


if __name__ == "__main__":
    sys.exit(main())



import copy
import hashlib
import json
import os
import urllib.request
import zipfile
from typing import Any, Dict

import generate_models as gm
import pytest
import xgboost
from xgboost import testing as tm
from xgboost.testing.updater import get_basescore


def run_model_param_check(name: str, config: Dict[str, Any]) -> None:
    assert config["learner"]["learner_model_param"]["num_feature"] == str(4)
    assert config["learner"]["learner_train_param"]["booster"] == "gbtree"

    booster = config["learner"]["gradient_booster"]
    assert booster["name"] == "gbtree"
    if name.find("1.0.0rc1") != -1:
        # There's no `num_parallel_tree` in the model parameter in 1.0 (it was a
        # configuration instead of a model parameter).
        return
    assert booster["gbtree_model_param"]["num_parallel_tree"] == str(gm.kForests)


def run_booster_check(booster: xgboost.Booster, name: str) -> None:
    config = json.loads(booster.save_config())
    run_model_param_check(name, config)
    n_rounds = get_n_rounds(name)
    if name.find("cls") != -1:
        assert len(booster.get_dump()) == gm.kForests * n_rounds * gm.kClasses
        base_score = get_basescore(config)
        assert isinstance(base_score, list)
        assert all(v == 0.5 for v in base_score)
        assert config["learner"]["learner_train_param"]["objective"] == "multi:softmax"
    elif name.find("logitraw") != -1:
        assert len(booster.get_dump()) == gm.kForests * n_rounds
        assert config["learner"]["learner_model_param"]["num_class"] == str(0)
        assert (
            config["learner"]["learner_train_param"]["objective"] == "binary:logitraw"
        )
    elif name.find("logit") != -1:
        assert len(booster.get_dump()) == gm.kForests * n_rounds
        assert config["learner"]["learner_model_param"]["num_class"] == str(0)
        assert (
            config["learner"]["learner_train_param"]["objective"] == "binary:logistic"
        )
    elif name.find("ltr") != -1:
        assert config["learner"]["learner_train_param"]["objective"] == "rank:ndcg"
    elif name.find("aft") != -1:
        assert config["learner"]["learner_train_param"]["objective"] == "survival:aft"
        assert (
            config["learner"]["objective"]["aft_loss_param"]["aft_loss_distribution"]
            == "normal"
        )
    else:
        assert name.find("reg") != -1
        assert len(booster.get_dump()) == gm.kForests * n_rounds
        assert get_basescore(config) == [0.5]
        assert (
            config["learner"]["learner_train_param"]["objective"] == "reg:squarederror"
        )


def get_n_rounds(name: str) -> int:
    if name.find("1.0.0rc1") != -1:
        n_rounds = 2
    else:
        n_rounds = gm.kRounds
    return n_rounds


def run_scikit_model_check(name: str, path: str) -> None:
    if name.find("reg") != -1:
        reg = xgboost.XGBRegressor()
        reg.load_model(path)
        config = json.loads(reg.get_booster().save_config())
        assert (
            config["learner"]["learner_train_param"]["objective"] == "reg:squarederror"
        )
        assert len(reg.get_booster().get_dump()) == get_n_rounds(name) * gm.kForests
        run_model_param_check(name, config)
    elif name.find("cls") != -1:
        cls = xgboost.XGBClassifier()
        cls.load_model(path)
        n_rounds = get_n_rounds(name)
        assert (
            len(cls.get_booster().get_dump()) == n_rounds * gm.kForests * gm.kClasses
        ), path
        config = json.loads(cls.get_booster().save_config())
        assert (
            config["learner"]["learner_train_param"]["objective"] == "multi:softprob"
        ), path
        run_model_param_check(name, config)
    elif name.find("ltr") != -1:
        ltr = xgboost.XGBRanker()
        ltr.load_model(path)
        assert len(ltr.get_booster().get_dump()) == get_n_rounds(name) * gm.kForests
        config = json.loads(ltr.get_booster().save_config())
        assert config["learner"]["learner_train_param"]["objective"] == "rank:ndcg"
        run_model_param_check(name, config)
    elif name.find("logitraw") != -1:
        logit = xgboost.XGBClassifier()
        logit.load_model(path)
        assert len(logit.get_booster().get_dump()) == get_n_rounds(name) * gm.kForests
        config = json.loads(logit.get_booster().save_config())
        assert (
            config["learner"]["learner_train_param"]["objective"] == "binary:logitraw"
        )
        run_model_param_check(name, config)
    elif name.find("logit") != -1:
        logit = xgboost.XGBClassifier()
        logit.load_model(path)
        assert len(logit.get_booster().get_dump()) == get_n_rounds(name) * gm.kForests
        config = json.loads(logit.get_booster().save_config())
        assert (
            config["learner"]["learner_train_param"]["objective"] == "binary:logistic"
        )
        run_model_param_check(name, config)
    else:
        assert False


def download(path: str) -> None:
    """Download the model files from S3."""
    zip_path, _ = urllib.request.urlretrieve(
        "https://xgboost-ci-jenkins-artifacts.s3-us-west-2"
        + ".amazonaws.com/xgboost_model_compatibility_tests-3.0.2.zip"
    )
    sha = "49d4d4db667a73590099dad9dca4f078532df05c5ea6e035ad4fa09596b1905a"
    with open(zip_path, "rb") as fd:
        digest = hashlib.file_digest(fd, "sha256")
        assert digest.hexdigest() == sha
    with zipfile.ZipFile(zip_path, "r") as z:
        z.extractall(path)


@pytest.mark.skipif(**tm.no_sklearn())
def test_model_compatibility() -> None:
    """Test model compatibility."""
    path = os.path.dirname(os.path.abspath(__file__))
    path = os.path.join(path, "models")

    if not os.path.exists(path):
        download(path)

    models = [
        os.path.join(root, f) for root, subdir, files in os.walk(path) for f in files
    ]
    assert len(models) == 54

    for path in models:
        name = os.path.basename(path)
        if name.startswith("xgboost-"):
            booster = xgboost.Booster(model_file=path)
            run_booster_check(booster, name)
            # Do full serialization.
            booster = copy.copy(booster)
            run_booster_check(booster, name)
        elif name.startswith("xgboost_scikit"):
            run_scikit_model_check(name, path)
        else:
            assert False



"""Tests for evaluation metrics."""

from typing import Dict, List, Optional

import numpy as np
import pytest

from ..compat import concat
from ..core import DMatrix, QuantileDMatrix, _parse_eval_str
from ..sklearn import XGBClassifier, XGBRanker
from ..training import train
from .utils import Device


def check_precision_score(  # pylint: disable=too-many-locals
    tree_method: str, device: Device
) -> None:
    """Test for precision with ranking and classification."""
    datasets = pytest.importorskip("sklearn.datasets")

    X, y = datasets.make_classification(
        n_samples=1024, n_features=4, n_classes=2, random_state=2023
    )
    qid = np.zeros(shape=y.shape)  # same group

    ltr = XGBRanker(n_estimators=2, tree_method=tree_method, device=device)
    ltr.fit(X, y, qid=qid)

    # re-generate so that XGBoost doesn't evaluate the result to 1.0
    X, y = datasets.make_classification(
        n_samples=512, n_features=4, n_classes=2, random_state=1994
    )

    ltr.set_params(eval_metric="pre@32")
    result = _parse_eval_str(ltr.get_booster().eval_set(evals=[(DMatrix(X, y), "Xy")]))
    score_0 = result[1][1]

    X_list = []
    y_list = []
    n_query_groups = 3
    q_list: List[np.ndarray] = []
    for i in range(n_query_groups):
        # same for all groups
        X, y = datasets.make_classification(
            n_samples=512, n_features=4, n_classes=2, random_state=1994
        )
        X_list.append(X)
        y_list.append(y)
        q = np.full(shape=y.shape, fill_value=i, dtype=np.uint64)
        q_list.append(q)

    qid = concat(q_list)
    X = concat(X_list)
    y = concat(y_list)

    result = _parse_eval_str(
        ltr.get_booster().eval_set(evals=[(DMatrix(X, y, qid=qid), "Xy")])
    )
    assert result[1][0].endswith("pre@32")
    score_1 = result[1][1]
    assert score_1 == score_0


def check_quantile_error(tree_method: str, device: Device) -> None:
    """Test for the `quantile` loss."""
    from sklearn.datasets import make_regression
    from sklearn.metrics import mean_pinball_loss

    rng = np.random.RandomState(19)
    # pylint: disable=unbalanced-tuple-unpacking
    X, y = make_regression(128, 3, random_state=rng)
    Xy = QuantileDMatrix(X, y)
    evals_result: Dict[str, Dict] = {}
    booster = train(
        {
            "tree_method": tree_method,
            "eval_metric": "quantile",
            "quantile_alpha": 0.3,
            "device": device,
        },
        Xy,
        evals=[(Xy, "Train")],
        evals_result=evals_result,
    )
    predt = booster.inplace_predict(X)
    loss = mean_pinball_loss(y, predt, alpha=0.3)
    np.testing.assert_allclose(evals_result["Train"]["quantile"][-1], loss)

    alpha = [0.25, 0.5, 0.75]
    booster = train(
        {
            "tree_method": tree_method,
            "eval_metric": "quantile",
            "quantile_alpha": alpha,
            "objective": "reg:quantileerror",
            "device": device,
        },
        Xy,
        evals=[(Xy, "Train")],
        evals_result=evals_result,
    )
    predt = booster.inplace_predict(X)
    loss = np.mean(
        [mean_pinball_loss(y, predt[:, i], alpha=alpha[i]) for i in range(3)]
    )
    np.testing.assert_allclose(evals_result["Train"]["quantile"][-1], loss)


def _expectile_loss(
    y: np.ndarray, predt: np.ndarray, alpha: float, weight: Optional[np.ndarray]
) -> float:
    diff = predt - y
    weight_scale = np.where(diff >= 0.0, 1.0 - alpha, alpha)
    loss = weight_scale * diff * diff
    if weight is None:
        return float(np.mean(loss))
    return float(np.sum(loss * weight) / np.sum(weight))


def _expectile_loss_multi(
    y: np.ndarray, predt: np.ndarray, alpha: np.ndarray, weight: Optional[np.ndarray]
) -> float:
    diff = predt - y[:, None]
    weight_scale = np.where(diff >= 0.0, 1.0 - alpha, alpha)
    loss = weight_scale * diff * diff
    if weight is None:
        return float(np.mean(loss))
    return float(np.sum(loss * weight[:, None]) / (np.sum(weight) * alpha.size))


def check_expectile_error(tree_method: str, device: Device) -> None:
    """Test for the `expectile` loss."""
    from sklearn.datasets import make_regression

    rng = np.random.RandomState(23)
    X, y = make_regression(128, 3, random_state=rng)
    Xy = DMatrix(X, y)
    evals_result: Dict[str, Dict] = {}
    booster = train(
        {
            "tree_method": tree_method,
            "eval_metric": "expectile",
            "expectile_alpha": 0.3,
            "device": device,
        },
        Xy,
        evals=[(Xy, "Train")],
        evals_result=evals_result,
    )
    predt = booster.inplace_predict(X)
    loss = _expectile_loss(y, predt, 0.3, None)
    np.testing.assert_allclose(evals_result["Train"]["expectile"][-1], loss)

    alpha = np.array([0.25, 0.5, 0.75])
    booster = train(
        {
            "tree_method": tree_method,
            "eval_metric": "expectile",
            "expectile_alpha": alpha,
            "objective": "reg:expectileerror",
            "device": device,
        },
        Xy,
        evals=[(Xy, "Train")],
        evals_result=evals_result,
    )
    predt = booster.inplace_predict(X)
    loss = _expectile_loss_multi(y, predt, alpha, None)
    np.testing.assert_allclose(evals_result["Train"]["expectile"][-1], loss)

    weights = rng.uniform(0.1, 1.0, size=y.shape[0])
    Xy_w = DMatrix(X, y, weight=weights)
    evals_result_w: Dict[str, Dict] = {}
    booster = train(
        {
            "tree_method": tree_method,
            "eval_metric": "expectile",
            "expectile_alpha": alpha,
            "objective": "reg:expectileerror",
            "device": device,
        },
        Xy_w,
        evals=[(Xy_w, "Train")],
        evals_result=evals_result_w,
    )
    predt = booster.inplace_predict(X)
    loss = _expectile_loss_multi(y, predt, alpha, weights)
    np.testing.assert_allclose(evals_result_w["Train"]["expectile"][-1], loss)


def run_roc_auc_binary(tree_method: str, n_samples: int, device: Device) -> None:
    """TestROC AUC metric on a binary classification problem."""
    from sklearn.datasets import make_classification
    from sklearn.metrics import roc_auc_score

    rng = np.random.RandomState(1994)
    n_features = 10

    X, y = make_classification(
        n_samples,
        n_features,
        n_informative=n_features,
        n_redundant=0,
        random_state=rng,
    )
    Xy = DMatrix(X, y)
    booster = train(
        {
            "tree_method": tree_method,
            "device": device,
            "eval_metric": "auc",
            "objective": "binary:logistic",
        },
        Xy,
        num_boost_round=1,
    )
    score = booster.predict(Xy)
    skl_auc = roc_auc_score(y, score)
    auc = float(booster.eval(Xy).split(":")[1])
    np.testing.assert_allclose(skl_auc, auc, rtol=1e-6)

    X = rng.randn(*X.shape)
    score = booster.predict(DMatrix(X))
    skl_auc = roc_auc_score(y, score)
    auc = float(booster.eval(DMatrix(X, y)).split(":")[1])
    np.testing.assert_allclose(skl_auc, auc, rtol=1e-6)


def run_pr_auc_multi(tree_method: str, device: Device) -> None:
    """Test for PR AUC metric on a multi-class classification problem."""
    from sklearn.datasets import make_classification

    X, y = make_classification(64, 16, n_informative=8, n_classes=3, random_state=1994)
    clf = XGBClassifier(
        tree_method=tree_method, n_estimators=1, eval_metric="aucpr", device=device
    )
    clf.fit(X, y, eval_set=[(X, y)])
    evals_result = clf.evals_result()["validation_0"]["aucpr"][-1]
    # No available implementation for comparison, just check that XGBoost converges
    # to 1.0
    clf = XGBClassifier(
        tree_method=tree_method, n_estimators=10, eval_metric="aucpr", device=device
    )
    clf.fit(X, y, eval_set=[(X, y)])
    evals_result = clf.evals_result()["validation_0"]["aucpr"][-1]
    np.testing.assert_allclose(1.0, evals_result, rtol=1e-2)


def run_roc_auc_multi(  # pylint: disable=too-many-locals
    tree_method: str, n_samples: int, weighted: bool, device: Device
) -> None:
    """Test for ROC AUC metric on a multi-class classification problem."""
    from sklearn.datasets import make_classification
    from sklearn.metrics import roc_auc_score

    rng = np.random.RandomState(1994)
    n_features = 10
    n_classes = 4

    X, y = make_classification(
        n_samples,
        n_features,
        n_informative=n_features,
        n_redundant=0,
        n_classes=n_classes,
        random_state=rng,
    )
    if weighted:
        weights = rng.randn(n_samples)
        weights -= weights.min()
        weights /= weights.max()
    else:
        weights = None

    Xy = DMatrix(X, y, weight=weights)
    booster = train(
        {
            "tree_method": tree_method,
            "eval_metric": "auc",
            "objective": "multi:softprob",
            "num_class": n_classes,
            "device": device,
        },
        Xy,
        num_boost_round=1,
    )
    score = booster.predict(Xy)
    skl_auc = roc_auc_score(
        y, score, average="weighted", sample_weight=weights, multi_class="ovr"
    )
    auc = float(booster.eval(Xy).split(":")[1])
    np.testing.assert_allclose(skl_auc, auc, rtol=1e-6)

    X = rng.randn(*X.shape)

    score = booster.predict(DMatrix(X, weight=weights))
    skl_auc = roc_auc_score(
        y, score, average="weighted", sample_weight=weights, multi_class="ovr"
    )
    auc = float(booster.eval(DMatrix(X, y, weight=weights)).split(":")[1])
    np.testing.assert_allclose(skl_auc, auc, rtol=1e-5)


def run_pr_auc_ltr(tree_method: str, device: Device) -> None:
    """Test for PR AUC metric on a ranking problem."""
    from sklearn.datasets import make_classification

    X, y = make_classification(128, 4, n_classes=2, random_state=1994)
    ltr = XGBRanker(
        tree_method=tree_method,
        n_estimators=16,
        objective="rank:pairwise",
        eval_metric="aucpr",
        device=device,
    )
    groups = np.array([32, 32, 64])
    ltr.fit(
        X,
        y,
        group=groups,
        eval_set=[(X, y)],
        eval_group=[groups],
    )
    results = ltr.evals_result()["validation_0"]["aucpr"]
    assert results[-1] >= 0.99


def run_pr_auc_binary(tree_method: str, device: Device) -> None:
    """Test for PR AUC metric on a binary classification problem."""
    from sklearn.datasets import make_classification
    from sklearn.metrics import auc, precision_recall_curve

    X, y = make_classification(128, 4, n_classes=2, random_state=1994)
    clf = XGBClassifier(
        tree_method=tree_method, n_estimators=1, eval_metric="aucpr", device=device
    )
    clf.fit(X, y, eval_set=[(X, y)])
    evals_result = clf.evals_result()["validation_0"]["aucpr"][-1]

    y_score = clf.predict_proba(X)[:, 1]  # get the positive column
    precision, recall, _ = precision_recall_curve(y, y_score)
    prauc = auc(recall, precision)
    # Interpolation results are slightly different from sklearn, but overall should
    # be similar.
    np.testing.assert_allclose(prauc, evals_result, rtol=1e-2)

    clf = XGBClassifier(
        tree_method=tree_method, n_estimators=10, eval_metric="aucpr", device=device
    )
    clf.fit(X, y, eval_set=[(X, y)])
    evals_result = clf.evals_result()["validation_0"]["aucpr"][-1]
    np.testing.assert_allclose(0.99, evals_result, rtol=1e-2)



import itertools
import warnings
from typing import Type

import numpy as np
import pytest
import scipy.sparse

import xgboost as xgb
from xgboost.testing.utils import predictor_equal


@pytest.mark.filterwarnings("error")
@pytest.mark.parametrize(
    "DMatrixT,CSR",
    [
        (m, n)
        for m, n in itertools.product(
            (xgb.DMatrix, xgb.QuantileDMatrix),
            (scipy.sparse.csr_matrix, scipy.sparse.csr_array),
        )
    ],
)
def test_csr(DMatrixT: Type[xgb.DMatrix], CSR: Type) -> None:
    with warnings.catch_warnings():
        indptr = np.array([0, 2, 3, 6])
        indices = np.array([0, 2, 2, 0, 1, 2])
        data = np.array([1, 2, 3, 4, 5, 6])
        X = CSR((data, indices, indptr), shape=(3, 3))
        dtrain = DMatrixT(X)
        assert dtrain.num_row() == 3
        assert dtrain.num_col() == 3
        assert dtrain.num_nonmissing() == data.size


@pytest.mark.filterwarnings("error")
@pytest.mark.parametrize(
    "DMatrixT,CSC",
    [
        (m, n)
        for m, n in itertools.product(
            (xgb.DMatrix, xgb.QuantileDMatrix),
            (scipy.sparse.csc_matrix, scipy.sparse.csc_array),
        )
    ],
)
def test_csc(DMatrixT: Type[xgb.DMatrix], CSC: Type) -> None:
    with warnings.catch_warnings():
        row = np.array([0, 2, 2, 0, 1, 2])
        col = np.array([0, 0, 1, 2, 2, 2])
        data = np.array([1, 2, 3, 4, 5, 6])
        X = CSC((data, (row, col)), shape=(3, 3))
        dtrain = DMatrixT(X)
        assert dtrain.num_row() == 3
        assert dtrain.num_col() == 3
        assert dtrain.num_nonmissing() == data.size

        indptr = np.array([0, 3, 5])
        data = np.array([0, 1, 2, 3, 4])
        row_idx = np.array([0, 1, 2, 0, 2])
        X = CSC((data, row_idx, indptr), shape=(3, 2))
        assert predictor_equal(DMatrixT(X.tocsr()), DMatrixT(X))


@pytest.mark.filterwarnings("error")
@pytest.mark.parametrize(
    "DMatrixT,COO",
    [
        (m, n)
        for m, n in itertools.product(
            (xgb.DMatrix, xgb.QuantileDMatrix),
            (scipy.sparse.coo_matrix, scipy.sparse.coo_array),
        )
    ],
)
def test_coo(DMatrixT: Type[xgb.DMatrix], COO: Type) -> None:
    with warnings.catch_warnings():
        row = np.array([0, 2, 2, 0, 1, 2])
        col = np.array([0, 0, 1, 2, 2, 2])
        data = np.array([1, 2, 3, 4, 5, 6])
        X = COO((data, (row, col)), shape=(3, 3))
        dtrain = DMatrixT(X)
        assert dtrain.num_row() == 3
        assert dtrain.num_col() == 3
        assert dtrain.num_nonmissing() == data.size

        assert predictor_equal(DMatrixT(X.tocsr()), DMatrixT(X))



"""
Demo for using `process_type` with `prune` and `refresh`
========================================================

Modifying existing trees is not a well established use for XGBoost, so feel free to
experiment.

"""

from urllib.error import HTTPError

import numpy as np
from sklearn.datasets import fetch_california_housing, make_regression

import xgboost as xgb


def main() -> None:
    n_rounds = 32

    try:
        X, y = fetch_california_housing(return_X_y=True)
    except HTTPError:
        # Use a synthetic dataset instead if we couldn't
        X, y = make_regression(n_samples=20640, n_features=8, random_state=1234)

    # Train a model first
    X_train = X[: X.shape[0] // 2]
    y_train = y[: y.shape[0] // 2]
    Xy = xgb.DMatrix(X_train, y_train)
    evals_result: xgb.callback.EvaluationMonitor.EvalsLog = {}
    booster = xgb.train(
        {"tree_method": "hist", "max_depth": 6, "device": "cuda"},
        Xy,
        num_boost_round=n_rounds,
        evals=[(Xy, "Train")],
        evals_result=evals_result,
    )
    SHAP = booster.predict(Xy, pred_contribs=True)

    # Refresh the leaf value and tree statistic
    X_refresh = X[X.shape[0] // 2 :]
    y_refresh = y[y.shape[0] // 2 :]
    Xy_refresh = xgb.DMatrix(X_refresh, y_refresh)
    # The model will adapt to other half of the data by changing leaf value (no change in
    # split condition) with refresh_leaf set to True.
    refresh_result: xgb.callback.EvaluationMonitor.EvalsLog = {}
    refreshed = xgb.train(
        {"process_type": "update", "updater": "refresh", "refresh_leaf": True},
        Xy_refresh,
        num_boost_round=n_rounds,
        xgb_model=booster,
        evals=[(Xy, "Original"), (Xy_refresh, "Train")],
        evals_result=refresh_result,
    )

    # Refresh the model without changing the leaf value, but tree statistic including
    # cover and weight are refreshed.
    refresh_result = {}
    refreshed = xgb.train(
        {"process_type": "update", "updater": "refresh", "refresh_leaf": False},
        Xy_refresh,
        num_boost_round=n_rounds,
        xgb_model=booster,
        evals=[(Xy, "Original"), (Xy_refresh, "Train")],
        evals_result=refresh_result,
    )
    # Without refreshing the leaf value, resulting trees should be the same with original
    # model except for accumulated statistic.  The rtol is for floating point error in
    # prediction.
    np.testing.assert_allclose(
        refresh_result["Original"]["rmse"], evals_result["Train"]["rmse"], rtol=1e-5
    )
    # But SHAP value is changed as cover in tree nodes are changed.
    refreshed_SHAP = refreshed.predict(Xy, pred_contribs=True)
    assert not np.allclose(SHAP, refreshed_SHAP, rtol=1e-3)

    # Prune the trees with smaller max_depth
    X_update = X_train
    y_update = y_train
    Xy_update = xgb.DMatrix(X_update, y_update)

    prune_result: xgb.callback.EvaluationMonitor.EvalsLog = {}
    pruned = xgb.train(
        {"process_type": "update", "updater": "prune", "max_depth": 2},
        Xy_update,
        num_boost_round=n_rounds,
        xgb_model=booster,
        evals=[(Xy, "Original"), (Xy_update, "Train")],
        evals_result=prune_result,
    )
    # Have a smaller model, but similar accuracy.
    np.testing.assert_allclose(
        np.array(prune_result["Original"]["rmse"]),
        np.array(prune_result["Train"]["rmse"]),
        atol=1e-5,
    )


if __name__ == "__main__":
    main()



from itertools import product

import pytest

from xgboost.testing.intercept import (
    run_adaptive,
    run_exp_family,
    run_init_estimation,
    run_logistic_degenerate,
)


def test_init_estimation() -> None:
    run_init_estimation("hist", "cuda")


@pytest.mark.parametrize(
    "tree_method,weighted", list(product(["approx", "hist"], [True, False]))
)
def test_adaptive(tree_method: str, weighted: bool) -> None:
    run_adaptive(tree_method, weighted, "cuda")


def test_exp_family() -> None:
    run_exp_family("cuda")


def test_logistic_degenerate() -> None:
    run_logistic_degenerate("cuda")



import numpy as np
import pandas as pd
import pytest

import xgboost as xgb
from xgboost import testing as tm
from xgboost.compat import is_dataframe
from xgboost.testing.data import run_base_margin_info

try:
    import modin.pandas as md
except ImportError:
    pass


pytestmark = pytest.mark.skipif(**tm.no_modin())


def test_type_check() -> None:
    df = md.DataFrame([[1, 2.0], [2, 3.0]], columns=["a", "b"])
    assert is_dataframe(df)
    assert is_dataframe(df.a)


class TestModin:
    @pytest.mark.xfail
    def test_modin(self) -> None:
        df = md.DataFrame([[1, 2., True], [2, 3., False]],
                          columns=['a', 'b', 'c'])
        dm = xgb.DMatrix(df, label=md.Series([1, 2]))
        assert dm.feature_names == ['a', 'b', 'c']
        assert dm.feature_types == ['int', 'float', 'i']
        assert dm.num_row() == 2
        assert dm.num_col() == 3
        np.testing.assert_array_equal(dm.get_label(), np.array([1, 2]))

        # overwrite feature_names and feature_types
        dm = xgb.DMatrix(df, label=md.Series([1, 2]),
                         feature_names=['x', 'y', 'z'],
                         feature_types=['q', 'q', 'q'])
        assert dm.feature_names == ['x', 'y', 'z']
        assert dm.feature_types == ['q', 'q', 'q']
        assert dm.num_row() == 2
        assert dm.num_col() == 3

        # incorrect dtypes
        df = md.DataFrame([[1, 2., 'x'], [2, 3., 'y']],
                          columns=['a', 'b', 'c'])
        with pytest.raises(ValueError):
            xgb.DMatrix(df)

        # numeric columns
        df = md.DataFrame([[1, 2., True], [2, 3., False]])
        dm = xgb.DMatrix(df, label=md.Series([1, 2]))
        assert dm.feature_names == ['0', '1', '2']
        assert dm.feature_types == ['int', 'float', 'i']
        assert dm.num_row() == 2
        assert dm.num_col() == 3
        np.testing.assert_array_equal(dm.get_label(), np.array([1, 2]))

        df = md.DataFrame([[1, 2., 1], [2, 3., 1]], columns=[4, 5, 6])
        dm = xgb.DMatrix(df, label=md.Series([1, 2]))
        assert dm.feature_names == ['4', '5', '6']
        assert dm.feature_types == ['int', 'float', 'int']
        assert dm.num_row() == 2
        assert dm.num_col() == 3

        df = md.DataFrame({'A': ['X', 'Y', 'Z'], 'B': [1, 2, 3]})
        dummies = md.get_dummies(df)
        #    B  A_X  A_Y  A_Z
        # 0  1    1    0    0
        # 1  2    0    1    0
        # 2  3    0    0    1
        result, _, _ = xgb.data._transform_pandas_df(dummies,
                                                     enable_categorical=False)
        exp = np.array([[1., 1., 0., 0.],
                        [2., 0., 1., 0.],
                        [3., 0., 0., 1.]]).T
        np.testing.assert_array_equal(result.columns, exp)
        dm = xgb.DMatrix(dummies)
        assert dm.feature_names == ['B', 'A_X', 'A_Y', 'A_Z']
        if int(pd.__version__[0]) >= 2:
            assert dm.feature_types == ["int", "i", "i", "i"]
        else:
            assert dm.feature_types == ["int", "int", "int", "int"]

        assert dm.num_row() == 3
        assert dm.num_col() == 4

        df = md.DataFrame({'A=1': [1, 2, 3], 'A=2': [4, 5, 6]})
        dm = xgb.DMatrix(df)
        assert dm.feature_names == ['A=1', 'A=2']
        assert dm.feature_types == ['int', 'int']
        assert dm.num_row() == 3
        assert dm.num_col() == 2

        df_int = md.DataFrame([[1, 1.1], [2, 2.2]], columns=[9, 10])
        dm_int = xgb.DMatrix(df_int)
        df_range = md.DataFrame([[1, 1.1], [2, 2.2]], columns=range(9, 11, 1))
        dm_range = xgb.DMatrix(df_range)
        assert dm_int.feature_names == ['9', '10']  # assert not "9 "
        assert dm_int.feature_names == dm_range.feature_names

        # test MultiIndex as columns
        df = md.DataFrame(
            [
                (1, 2, 3, 4, 5, 6),
                (6, 5, 4, 3, 2, 1)
            ],
            columns=md.MultiIndex.from_tuples((
                ('a', 1), ('a', 2), ('a', 3),
                ('b', 1), ('b', 2), ('b', 3),
            ))
        )
        dm = xgb.DMatrix(df)
        assert dm.feature_names == ['a 1', 'a 2', 'a 3', 'b 1', 'b 2', 'b 3']
        assert dm.feature_types == ['int', 'int', 'int', 'int', 'int', 'int']
        assert dm.num_row() == 2
        assert dm.num_col() == 6

    def test_modin_label(self):
        # label must be a single column
        df = md.DataFrame({"A": ["X", "Y", "Z"], "B": [1, 2, 3]})
        with pytest.raises(ValueError):
            xgb.data._transform_pandas_df(df, False, None, None, "label")

        # label must be supported dtype
        df = md.DataFrame({"A": np.array(["a", "b", "c"], dtype=object)})
        with pytest.raises(ValueError):
            xgb.data._transform_pandas_df(df, False, None, None, "label")

        df = md.DataFrame({"A": np.array([1, 2, 3], dtype=int)})
        result, _, _ = xgb.data._transform_pandas_df(
            df, False, None, None, "label"
        )
        np.testing.assert_array_equal(
            np.stack(result.columns, axis=1),
            np.array([[1.0], [2.0], [3.0]], dtype=float),
        )
        dm = xgb.DMatrix(np.random.randn(3, 2), label=df)
        assert dm.num_row() == 3
        assert dm.num_col() == 2

    def test_modin_weight(self):
        kRows = 32
        kCols = 8

        X = np.random.randn(kRows, kCols)
        y = np.random.randn(kRows)
        w = np.random.uniform(size=kRows).astype(np.float32)
        w_pd = md.DataFrame(w)
        data = xgb.DMatrix(X, y, w_pd)

        assert data.num_row() == kRows
        assert data.num_col() == kCols

        np.testing.assert_array_equal(data.get_weight(), w)

    def test_base_margin(self):
        run_base_margin_info(md.DataFrame, xgb.DMatrix, "cpu")
